import datetime
import stat

import pytest

import salt.exceptions

try:
    import cryptography
    from cryptography.hazmat.primitives import serialization

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False

try:
    import bcrypt  # pylint: disable=unused-import

    HAS_BCRYPT = True
except ImportError:
    HAS_BCRYPT = False

CRYPTOGRAPHY_VERSION = tuple(int(x) for x in cryptography.__version__.split("."))

pytestmark = [
    pytest.mark.skipif(HAS_LIBS is False, reason="Needs cryptography library"),
]


@pytest.fixture(scope="module")
def minion_config_overrides():
    return {
        "ssh_signing_policies": {
            "testhostpolicy": {
                "cert_type": "host",
                "valid_principals": ["from_host_signing_policy"],
                "allowed_extensions": [],
                "allowed_critical_options": [],
            },
            "testuserpolicy": {
                "cert_type": "user",
                "valid_principals": ["from_user_signing_policy"],
                "critical_options": {
                    "force-command": "echo hi",
                    "no-port-forwarding": True,
                    "verify-required": False,
                },
                "extensions": {
                    "permit-X11-forwarding": True,
                    "permit-agent-forwarding": True,
                    "permit-port-forwarding": True,
                    "permit-pty": True,
                    "permit-user-rc": False,
                },
            },
            "testprincipalspolicy": {
                "cert_type": "host",
                "valid_principals": ["a", "b", "c"],
            },
        },
    }


@pytest.fixture
def ssh(modules):
    yield modules.ssh_pki


@pytest.fixture
def ca_key():
    return """\
-----BEGIN OPENSSH PRIVATE KEY-----
b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAABFwAAAAdzc2gtcnNhAAAA
AwEAAQAAAQEAmRii0ZT9KoL3aC9X7I8uQmcVGa4G7qOlyInNxGgKPC3s5+kshIVTMQfugd1irFT7
crF6Kamigq0S3uNBzpq/IxQBIRq8n7nHHFxiTgi+SK61XXMDdlbLsdXZxh7TkTMYPp32nmKUFDdd
xAVQCGwFvzkV2tbtxpiRO6E/60b22Lpt3qS0JdSgSYp9q0IFv2iNR+Yp+Kovx/I8+RySdhgecpRr
16Qul+DqvEqNtXvNJv4uA1nh34GCKBvNa8JekCYhGWpiXKvDaJREEFfVtuL0YrgDCS6JmbqMreYq
zsl8+QN9gEkt4JjDTQkEGUI7a4ZY+m69w2eN5eT0DebaNWGzkQAAA7gCJksAAiZLAAAAAAdzc2gt
cnNhAAABAQCZGKLRlP0qgvdoL1fsjy5CZxUZrgbuo6XIic3EaAo8Lezn6SyEhVMxB+6B3WKsVPty
sXopqaKCrRLe40HOmr8jFAEhGryfucccXGJOCL5IrrVdcwN2Vsux1dnGHtORMxg+nfaeYpQUN13E
BVAIbAW/ORXa1u3GmJE7oT/rRvbYum3epLQl1KBJin2rQgW/aI1H5in4qi/H8jz5HJJ2GB5ylGvX
pC6X4Oq8So21e80m/i4DWeHfgYIoG81rwl6QJiEZamJcq8NolEQQV9W24vRiuAMJLomZuoyt5irO
yXz5A32ASS3gmMNNCQQZQjtrhlj6br3DZ43l5PQN5to1YbORAAAAAwEAAQAAAQAcJGIxonCTMvXl
qeZAruUzAZ3oVYwiq+Rao7I2a2WOQGbvDnbHeXacabfXGWn9AbYjFCq/o9YirUvtutqq7tk5yoCW
pEKOHelS9kx/ya2o0Ky4G99EDpps+0GH4LzFUR4gzIq7/KT5vl+3G77lfW3lA8pXqvCUdBEmY/LK
/gV8OyffLxdQMIYCucpVpJq/wSiIYdd3DVK2TvlmUsR0DnvF5zo8/b7nvkALElXdA5G9NUL+k+XZ
0iDiWB0Ezokg5/VokhhhP8NcgiHeO00XHE6AAQ8rbUq9cwRgn0oqyC4ZYMc68MZOQhn2GKcYTRxG
YbGC6xoIUWrvaQq1/L4NEDIRAAAAgB9FHj/pVd/7HDjkGF9YapBu4kvuv9AxGvOVOWWHIk0pQl/L
vOEFj0gz4xl+BEdIDpX0RVS1p0wPD3h8uBKWQ5q2t4wOIzGidfaL+u/NHo4MPVC+cOCfgsRU7sy2
vN9kOjNpP7C3QTOzojq2vJ0zkJt+nZ3IA6OxyvS1Ebc0j88qAAAAgQDVjG+C+YyrYy9m5KBoj5T4
UEjNzgCr9t6ANSHObQjQxafRfeB4gT7vv65xna7Q4I8ATm2wITty1LETLNWbc2B4W5Mir8CnaUw0
P9Eyxql7VaTrHU1M3ee50Yquz1UEFLat1JXE/jMc2vs7s/W5Vd8uitom3uK8IkY/v3XOC4nk1QAA
AIEAt4fCQs36AcsYngt7oboKBMCnaiFWfupvZTAC86kd7VcuF74r7IrC4U9Oohwl1l6jXU90p0Mm
qJJROD1G3S1IFenRXQ6aIizaO7WkFRBeEuUtHwI0MbBq2cDjVzvRsu9qnLgMCeVMuBmj7pVG+Isf
zgx0+GaEc+f1RlC0qh6eIc0AAAAAAQID
-----END OPENSSH PRIVATE KEY-----"""


@pytest.fixture
def ca_key_enc():
    return """\
-----BEGIN OPENSSH PRIVATE KEY-----
b3BlbnNzaC1rZXktdjEAAAAACmFlczI1Ni1jdHIAAAAGYmNyeXB0AAAAGAAAABAo0CeCU7RU21xt
OjRqiR2WAAAAEAAAAAEAAAEXAAAAB3NzaC1yc2EAAAADAQABAAABAQCZGKLRlP0qgvdoL1fsjy5C
ZxUZrgbuo6XIic3EaAo8Lezn6SyEhVMxB+6B3WKsVPtysXopqaKCrRLe40HOmr8jFAEhGryfuccc
XGJOCL5IrrVdcwN2Vsux1dnGHtORMxg+nfaeYpQUN13EBVAIbAW/ORXa1u3GmJE7oT/rRvbYum3e
pLQl1KBJin2rQgW/aI1H5in4qi/H8jz5HJJ2GB5ylGvXpC6X4Oq8So21e80m/i4DWeHfgYIoG81r
wl6QJiEZamJcq8NolEQQV9W24vRiuAMJLomZuoyt5irOyXz5A32ASS3gmMNNCQQZQjtrhlj6br3D
Z43l5PQN5to1YbORAAADwIahmMW2HPaQ+r0WV3eT5lkbQjeTIkOo47PexilFof9ZYoOOk3AjcgYP
1SEhq4BXnMnpTVbRUkmzR7+od7tXrqVC9Eo3NuOwyjZwRJuHB+zPKpYqZ9SPOMRVYxx75yUmXnCu
Hsm8zT9F1izm/KQOyoSlbcWK3uQSSvYHYAWku/5bc6+fCPbgFYlB1Bk+cxjeGOPD+qDWQd4IKXj/
W/KCE8ncjQIZkrX1fgPGJsxCJwNUwYpJUad2+8XaZJIAIur36tlyNw+1JmQHSGMhfLW1P+tDXAIc
JJ+KpK/uVhYX5uOdER3/trwa6YQBHuUjo3sGrfYYTprd+WH0PbN2QJ9Vi2m5gJLY7CjCod7ecjbJ
ntTN9xucis+I7eZZPHhd8MO9wNQkXRwy07ZPawExlv2EVDUsofHQxiuHkvPZHNq4eFDz906xIk0v
n0X9IBrVtjg+bvnhl9XG86Ay5NAD7vCquqEqZAAHet24oi7yJmOkAJVA/0PaPzaALfp1a7noFFud
FuC9PmmdTe0gL1m1pQki1fIFMk2dxB83pQiBJThn9C3KhbrtsCN00kWxb3Ezv2mFPNeEs3Q9W7r5
mHVuJfVDusXAdXlpp+EcQueZW1r0Qtnf0UbH57LiTt6S1MUEYn3jgIE/ynHOPTrv0RgE8udCq8/g
1gUcZwL8BLJVCu+Dllj4/l/s/E3BzaTCzmc48toAdbeoZ1+nzXBCez1i8QQZ57gQK7hVfNCqCUD9
yNw84panB1lkxvBxFmN05dKxeBCGY0Bpxnl5xb2ND+rASXCELSEnHxt+9FVEdjrlpyg+KUnuBort
JmmK/VKmeW2yk+7kHDTefKpRiK7PNF1gUko15gzegq0YEGDewo/RT/c9BaYHZougg5HBjT6bUHaN
6x2Jc75z5ILXaeLrsavIyYuhMwpBrMYSPm/xNZvBxPiM+AyhCq2YAgS0eV3Asd4h7HB90NZHbwjc
nvmCLpiZtn269dM1wZdYUGaPCf6Gmj5OojJVRhgaXD8SHOzvb5HqQnco3e+KiVrFCGvfn/S8CTnv
CsQx2szAQcgkO16xsKZC8p1DAJSH+palgYRY2gMjXDzQN57RBGZPd2wwdSUc/iYXjUnXz782Nb/W
a7MutFQiFnqr0WUXscoenXsxDYjU8D81hbqkxVXzX/uwOmux0ilPwX1RlvjB2TuZ0YjDBtdprOYf
tXpvXY2joVcZBnblNTC+AMZQAza2xRdT5/VdNK7b6cdjQ00lo+OMRNeWLcjEL6JI8siEAywW/h69
Dft9BkNgrA==
-----END OPENSSH PRIVATE KEY-----"""


@pytest.fixture
def ca_pub():
    return (
        "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQCZGKLRlP0qgvdoL1fsjy5CZxUZrg"
        "buo6XIic3EaAo8Lezn6SyEhVMxB+6B3WKsVPtysXopqaKCrRLe40HOmr8jFAEhGryf"
        "ucccXGJOCL5IrrVdcwN2Vsux1dnGHtORMxg+nfaeYpQUN13EBVAIbAW/ORXa1u3GmJ"
        "E7oT/rRvbYum3epLQl1KBJin2rQgW/aI1H5in4qi/H8jz5HJJ2GB5ylGvXpC6X4Oq8"
        "So21e80m/i4DWeHfgYIoG81rwl6QJiEZamJcq8NolEQQV9W24vRiuAMJLomZuoyt5i"
        "rOyXz5A32ASS3gmMNNCQQZQjtrhlj6br3DZ43l5PQN5to1YbOR"
    )


@pytest.fixture
def rsa_privkey():
    return """\
-----BEGIN OPENSSH PRIVATE KEY-----
b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAABFwAAAAdzc2gtcnNhAAAA
AwEAAQAAAQEAt11Mkho+ChXlHC33p5I+K0r2wldrHV48T+5fHtuOrHNsaR01perekmDFQMvnhgDh
P/ibFENrFgo/yg2MHh/SldgC6fQFqZr89uJ+/E63oZ3+EIvUdQWiN8jiPkgKTTOlMXh2DjuPPa6U
C8hEZwnE5lg+/UN5BjiivZCbO0/Ll8/MXlPvSk3tdIao3twHJ4WwKYPotXwZ6LUM7HFloC/yPLfj
/AQhz6GO1fT5wLobDefB8Wpg2bWPO0yLvKlDGbZWzn2xy7deP2WrjqUsgTM9PaUKLIVBEgB/Apw2
FUkNkM6lHGM1bBAsiY5L3ISqBIku1us2TvwRZRGKDQYYIm3mbwAAA7gckLhOHJC4TgAAAAdzc2gt
cnNhAAABAQC3XUySGj4KFeUcLfenkj4rSvbCV2sdXjxP7l8e246sc2xpHTWl6t6SYMVAy+eGAOE/
+JsUQ2sWCj/KDYweH9KV2ALp9AWpmvz24n78Trehnf4Qi9R1BaI3yOI+SApNM6UxeHYOO489rpQL
yERnCcTmWD79Q3kGOKK9kJs7T8uXz8xeU+9KTe10hqje3AcnhbApg+i1fBnotQzscWWgL/I8t+P8
BCHPoY7V9PnAuhsN58HxamDZtY87TIu8qUMZtlbOfbHLt14/ZauOpSyBMz09pQoshUESAH8CnDYV
SQ2QzqUcYzVsECyJjkvchKoEiS7W6zZO/BFlEYoNBhgibeZvAAAAAwEAAQAAAQA6mf3Lv8lODdtQ
GzzH+EwYJ/ge+jNIioJ6BTOvN/osESN3oJcxtJN3cqf34nLW99cFS928JyPFQndKivPZ+M+jhgrA
XHWm09q+yHpPBpVXeJfnD9lRoQBMFc6AmyN3sua7ncUVHWHVE8NK1LFPwOaFu3Q+Gt9F5rnUHbAO
z5zAomT10HfQd9yeDCdywAFsPklwbzmzv72ps3+CSG+ppRm+oh1GYYe7zWmX4SBPuTJbx9BTcuWd
yQTvSUIamP/22gxoNMzNBkYu/WeXNpB5iHGazNMAIEdyJ9x8VimnUIyZeW/hIPXt5MUW/WfFOE8Z
zzQ/+MTPF+uSfBC54tJQUvrxAAAAgD4wQaNkRY5jNZ3c4SUsrzwDDdbWH0SbTGLcnCWpd809apTN
adIBI+vt74kZCYafOqOOldyeU3Akw/fc9+BoBrR62boAYTSWlWNFqsNcE0AjEJbO0fIm2ngr8Z+o
vFp7fIyWQaIklmV0eyetAOz4InF+xC7SJqH2qm38kQvc0bdWAAAAgQDlg1w1HnyiBmooml3Ql5ln
uq+Fv8XZ/xxfVxRY5MCugVAR9eLQcMUJ/s3o44VEu1N3GAe1Dge3HYSQMPtZHIIe2AG38u1wHR+t
fFAQswoOD1Fn3O69CIG8kjzqdPLo12E+0+gJI5ki6lgAGEkVbQW4GsqVm4jBF0SPjUTfN5FRlwAA
AIEAzIaKrsfUUAPBCLuYeiZGfSCiXtdVZ+Yo8ADMwHA7iMSWG6ushLNwcV3PqRjVbW3xfM+9OghX
/zM6sratQmVpqhTBwQnyXBJp5SFT2n/XFYwkKhERBukBB/uEcI6M4HYhP8Z/N8lE7PNuY0ylnmrT
Cy/i60jGoFHQaROUPbs2/OkAAAAAAQID
-----END OPENSSH PRIVATE KEY-----"""


@pytest.fixture
def rsa_privkey_enc():
    return """\
-----BEGIN OPENSSH PRIVATE KEY-----
b3BlbnNzaC1rZXktdjEAAAAACmFlczI1Ni1jdHIAAAAGYmNyeXB0AAAAGAAAABA8ZLxKjCh9dRM4
7AmjoqGqAAAAEAAAAAEAAAEXAAAAB3NzaC1yc2EAAAADAQABAAABAQC3XUySGj4KFeUcLfenkj4r
SvbCV2sdXjxP7l8e246sc2xpHTWl6t6SYMVAy+eGAOE/+JsUQ2sWCj/KDYweH9KV2ALp9AWpmvz2
4n78Trehnf4Qi9R1BaI3yOI+SApNM6UxeHYOO489rpQLyERnCcTmWD79Q3kGOKK9kJs7T8uXz8xe
U+9KTe10hqje3AcnhbApg+i1fBnotQzscWWgL/I8t+P8BCHPoY7V9PnAuhsN58HxamDZtY87TIu8
qUMZtlbOfbHLt14/ZauOpSyBMz09pQoshUESAH8CnDYVSQ2QzqUcYzVsECyJjkvchKoEiS7W6zZO
/BFlEYoNBhgibeZvAAADwDiKlu/+uo3XsapQBgGxZOSTHY1bqXZG1qh1EaE0DxXl0RQApJkZCutV
2wWN8VaCNbtdcLgDTqC8WMHd/QzseJBWabnIUHZXYsCbU9BmcfjUhCQ0kyN2zSb4YmsM73PqMX6I
5usU0nVAV/UhBSMg7tIzMVBQzYLYl+eypW9usp5HYb8dw/vMhQY+a1AYqIpe9vZlPohso8U2Q9bL
LiMNjMSSWcgh2b3i/LbixzHTGSibF9/izntSem/jxnoZ27R1r97KcuxPF3uKfPlrlpsmzNqhVi0d
4YWMttG3EUrpsGHUG5Ylp8bXnfMfmFkmkiEGjU/Gf30tiKlFznQ4G9luYn7h4Rs8vx9U7Xolhz22
yqJ0FxvvY9HYIhK6WTqOdPKE7O1cy1cn9Rne6RG4ALmETjbVh1dMCIKBFaIYztWIoT87XJrH3piD
xRlMQog7Cim/a2zd1l/G+iyEDBoVUzWx5WNJBiJgt9ZbYEJz1cJiWFDJhzWO1mFXYzNaJdfUdL43
xZjU+sxCxXuV+ps6nBjz9CmnI7hkNAIoa9lBNUo8uWds2DnNm0QU48bliVtxcLcjAdpZev5H1Hfx
bPgHDrja3F0fbtJaoadIn/xh8c0CTtOCKgjS1up9YnB2AL+bFoX0vcl13kRvuxnpUDaxQ2xehETR
SBupMQmGcdSSOUhLIF4/QJFijwMHNhUcIJqgzhT256z/dEKxtofi4Q9TiRjDT4uENO9Uy/FbZWAI
zaW3TLPZAnTLTIishfCOOm0yb9XmuxM4hLgGpqAHSxEvSqMXS4fe+BIdj2i5cOsB/N3//OE1yQ1M
qeMqTyvXEbK+PAH65rdhE6xyeJFeKV+fil7dyZX41R6FUtZ9ls3AXcmqFIq463n7MvUGkl8zqMvz
EzH8oCQRJ1Taly9JsLgCd7UbHpokeQeMYKyd5+muYvBu0qX2HPD0exJ5uJSxSqq+0NrlDKmY1INb
/XcPBjxVtKsXoTRfh/vs+TcLxwYbAWymWCdX3k4maa7oJ3ZOc0L5/ypeVUaecW4PcyXlYzEAfIwK
nFAirfDVSpjBochJfNBbhCgmtdmq3nUh8CI7AExJmVrWvEbJ6klJMVet0m50mfqyhBI5idw2YfKs
qxOoAdt3IZIxmGG+pyDeeFB8X9O5RF1BVnFsqNPHEqdRtVD7qPSYBaZDhtgFVJbstwW3+///nyLP
5B5yBiGVm0aFXpFcVeCHYX817crRC+0h/HA5fECosGRMZFZEAJh5CHivuKNvhseKdnWXLwkfIpuO
Ojnp0x6Zow==
-----END OPENSSH PRIVATE KEY-----"""


@pytest.fixture
def rsa_pubkey():
    return (
        "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC3XUySGj4KFeUcLfenkj4rSvbCV2s"
        "dXjxP7l8e246sc2xpHTWl6t6SYMVAy+eGAOE/+JsUQ2sWCj/KDYweH9KV2ALp9AWpmv"
        "z24n78Trehnf4Qi9R1BaI3yOI+SApNM6UxeHYOO489rpQLyERnCcTmWD79Q3kGOKK9k"
        "Js7T8uXz8xeU+9KTe10hqje3AcnhbApg+i1fBnotQzscWWgL/I8t+P8BCHPoY7V9PnA"
        "uhsN58HxamDZtY87TIu8qUMZtlbOfbHLt14/ZauOpSyBMz09pQoshUESAH8CnDYVSQ2"
        "QzqUcYzVsECyJjkvchKoEiS7W6zZO/BFlEYoNBhgibeZv"
    )


@pytest.fixture
def ec_privkey():
    return """\
-----BEGIN OPENSSH PRIVATE KEY-----
b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAAAaAAAABNlY2RzYS1zaGEy
LW5pc3RwMjU2AAAACG5pc3RwMjU2AAAAQQScOqN2HTar84e+l3Er4Ti0ZsY3nbR9RkRsgZb0Flie
lc8SN/zIHSLroaJ21ofSqfu+mazGpGFWkNo34zSbBW5+AAAAoEaJqOBGiajgAAAAE2VjZHNhLXNo
YTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBJw6o3YdNqvzh76XcSvhOLRmxjedtH1GRGyBlvQW
WJ6VzxI3/MgdIuuhonbWh9Kp+76ZrMakYVaQ2jfjNJsFbn4AAAAhAL7DzNqGQYRNLOeXUt/t+DFz
R4+26CwTk8SDLHiIt2dpAAAAAAECAwQFBgc=
-----END OPENSSH PRIVATE KEY-----"""


@pytest.fixture
def ec_pubkey():
    return (
        "ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNT"
        "YAAABBBJw6o3YdNqvzh76XcSvhOLRmxjedtH1GRGyBlvQWWJ6VzxI3/MgdIuuhonbW"
        "h9Kp+76ZrMakYVaQ2jfjNJsFbn4="
    )


@pytest.fixture
def ed25519_privkey():
    return """\
-----BEGIN OPENSSH PRIVATE KEY-----
b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAAAMwAAAAtzc2gtZWQyNTUx
OQAAACDRCIjV5alJaolYOvZSZMzH8gVgGp7LsKlum8O86Me5EQAAAIidIDmZnSA5mQAAAAtzc2gt
ZWQyNTUxOQAAACDRCIjV5alJaolYOvZSZMzH8gVgGp7LsKlum8O86Me5EQAAAEBc5nqUPbCxESlT
vH72xFJW6XGhm9Sm4CbrmcdfhucVdtEIiNXlqUlqiVg69lJkzMfyBWAansuwqW6bw7zox7kRAAAA
AAECAwQF
-----END OPENSSH PRIVATE KEY-----"""


@pytest.fixture
def ed25519_pubkey():
    return "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAINEIiNXlqUlqiVg69lJkzMfyBWAansuwqW6bw7zox7kR"


@pytest.fixture
def cert_exts():
    return (
        "ssh-rsa-cert-v01@openssh.com AAAAHHNzaC1yc2EtY2VydC12MDFAb3BlbnNzaC5"
        "jb20AAAAgAutZE2GohcxGXD1rmPbhaLcN5uLcWjBk0KQIXOcDr5oAAAADAQABAAABAQC"
        "3XUySGj4KFeUcLfenkj4rSvbCV2sdXjxP7l8e246sc2xpHTWl6t6SYMVAy+eGAOE/+Js"
        "UQ2sWCj/KDYweH9KV2ALp9AWpmvz24n78Trehnf4Qi9R1BaI3yOI+SApNM6UxeHYOO48"
        "9rpQLyERnCcTmWD79Q3kGOKK9kJs7T8uXz8xeU+9KTe10hqje3AcnhbApg+i1fBnotQz"
        "scWWgL/I8t+P8BCHPoY7V9PnAuhsN58HxamDZtY87TIu8qUMZtlbOfbHLt14/ZauOpSy"
        "BMz09pQoshUESAH8CnDYVSQ2QzqUcYzVsECyJjkvchKoEiS7W6zZO/BFlEYoNBhgibeZ"
        "vA2htsw1e8lQAAAABAAAADXNhbHRzdGFja3Rlc3QAAAARAAAABHNhbHQAAAAFc3RhY2s"
        "AAAAAZJvNIgAAAABknR6iAAAAOgAAAA1mb3JjZS1jb21tYW5kAAAACwAAAAdlY2hvIGh"
        "pAAAAEm5vLXBvcnQtZm9yd2FyZGluZwAAAAAAAAA8AAAAFXBlcm1pdC1YMTEtZm9yd2F"
        "yZGluZwAAAAAAAAAXcGVybWl0LWFnZW50LWZvcndhcmRpbmcAAAAAAAAAAAAAARcAAAA"
        "Hc3NoLXJzYQAAAAMBAAEAAAEBAJkYotGU/SqC92gvV+yPLkJnFRmuBu6jpciJzcRoCjw"
        "t7OfpLISFUzEH7oHdYqxU+3KxeimpooKtEt7jQc6avyMUASEavJ+5xxxcYk4IvkiutV1"
        "zA3ZWy7HV2cYe05EzGD6d9p5ilBQ3XcQFUAhsBb85FdrW7caYkTuhP+tG9ti6bd6ktCX"
        "UoEmKfatCBb9ojUfmKfiqL8fyPPkcknYYHnKUa9ekLpfg6rxKjbV7zSb+LgNZ4d+Bgig"
        "bzWvCXpAmIRlqYlyrw2iURBBX1bbi9GK4AwkuiZm6jK3mKs7JfPkDfYBJLeCYw00JBBl"
        "CO2uGWPpuvcNnjeXk9A3m2jVhs5EAAAEUAAAADHJzYS1zaGEyLTUxMgAAAQAUJpzjwKC"
        "Ib2oHLo4q6v6qN37ssRmprwtAVJaV9FmAgczroMyWiL3P5O7U4TLPzluODVRSzpI1K5j"
        "oeIheQQ24npcErmxVj3xkk4uFsaTgyAamQfnudlseltmNfzqYX6LStFco27bBM6x1V3p"
        "TPbOfd8m1g1JZEH8yGCt8J4zpVNs6NWTJ862PqdlQKYaZcfEdDyuW0AemSug7YXBe6XU"
        "6F5v55T8W4sGy3yLn9HWMFyvdtnH91Y8NzsC3JgQfuImPDmdH0hjg0Fr/9579EUEIiq3"
        "IHN7S4e2vrj7lIreSxPeKnkQNvrjmrPojLD9atastKTalIriT7mtfqzOTih7V"
    )


@pytest.fixture
def cert_exts_read():
    return {
        "cert_type": "user",
        "key_id": "saltstacktest",
        "key_size": 2048,
        "key_type": "rsa",
        "serial_number": "03:68:6D:B3:0D:5E:F2:54",
        "issuer_public_key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQCZGKLRlP0qgvdoL1fsjy5CZxUZrgbuo6XIic3EaAo8Lezn6SyEhVMxB+6B3WKsVPtysXopqaKCrRLe40HOmr8jFAEhGryfucccXGJOCL5IrrVdcwN2Vsux1dnGHtORMxg+nfaeYpQUN13EBVAIbAW/ORXa1u3GmJE7oT/rRvbYum3epLQl1KBJin2rQgW/aI1H5in4qi/H8jz5HJJ2GB5ylGvXpC6X4Oq8So21e80m/i4DWeHfgYIoG81rwl6QJiEZamJcq8NolEQQV9W24vRiuAMJLomZuoyt5irOyXz5A32ASS3gmMNNCQQZQjtrhlj6br3DZ43l5PQN5to1YbOR",
        "issuer_public_key_fingerprints": {
            "md5": "e1:6c:9c:1b:33:22:dd:aa:6f:7e:07:80:f0:92:95:58",
            "sha256": "AkjrJzyLdJXTj+36AVeT7UeK0efm1OMPu8ovT3UNHU0=",
        },
        "not_before": "2023-06-28 06:03:14",
        "not_after": "2023-06-29 06:03:14",
        "public_key": "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC3XUySGj4KFeUcLfenkj4rSvbCV2sdXjxP7l8e246sc2xpHTWl6t6SYMVAy+eGAOE/+JsUQ2sWCj/KDYweH9KV2ALp9AWpmvz24n78Trehnf4Qi9R1BaI3yOI+SApNM6UxeHYOO489rpQLyERnCcTmWD79Q3kGOKK9kJs7T8uXz8xeU+9KTe10hqje3AcnhbApg+i1fBnotQzscWWgL/I8t+P8BCHPoY7V9PnAuhsN58HxamDZtY87TIu8qUMZtlbOfbHLt14/ZauOpSyBMz09pQoshUESAH8CnDYVSQ2QzqUcYzVsECyJjkvchKoEiS7W6zZO/BFlEYoNBhgibeZv",
        "public_key_fingerprints": {
            "md5": "b0:c2:ae:34:55:8a:27:d5:87:cd:a7:3c:54:ca:59:6d",
            "sha256": "24C6IQIPizP36/mQRNth/L3rTHHM6JgpppKQj7U+xAA=",
        },
        "critical_options": {"force-command": "echo hi", "no-port-forwarding": True},
        "extensions": {"permit-X11-forwarding": True, "permit-agent-forwarding": True},
        "valid_principals": ["salt", "stack"],
    }


def _assert_cert_basic(res, cert_type, pubkey, ca_pubkey, valid_principals=None):
    cert = _get_cert(res)
    if isinstance(cert_type, str):
        cert_type = getattr(serialization.SSHCertificateType, cert_type.upper())
    assert cert.type is cert_type
    assert cert.key_id == b"success"
    assert _encode_pubkey(cert.public_key()) == pubkey
    assert _encode_pubkey(cert.signature_key()) == ca_pubkey
    assert cert.valid_principals == (valid_principals or [])
    return cert


@pytest.mark.parametrize(
    "algo",
    [
        "rsa",
        "ec",
        pytest.param("ed25519", marks=pytest.mark.skip_on_fips_enabled_platform),
    ],
)
@pytest.mark.parametrize("cert_type", ["user", "host"])
def test_create_certificate_from_privkey(ssh, cert_type, ca_key, ca_pub, algo, request):
    privkey = request.getfixturevalue(f"{algo}_privkey")
    pubkey = request.getfixturevalue(f"{algo}_pubkey")
    res = ssh.create_certificate(
        cert_type=cert_type,
        signing_private_key=ca_key,
        private_key=privkey,
        key_id="success",
        all_principals=True,
    )
    _assert_cert_basic(res, cert_type, pubkey, ca_pub)


@pytest.mark.skipif(
    HAS_BCRYPT is False, reason="Encrypted keys require the bcrypt library"
)
@pytest.mark.parametrize("cert_type", ["user", "host"])
def test_create_certificate_from_encrypted_privkey(
    ssh, ca_key, ca_pub, rsa_privkey_enc, rsa_pubkey, cert_type
):
    res = ssh.create_certificate(
        cert_type=cert_type,
        signing_private_key=ca_key,
        private_key=rsa_privkey_enc,
        private_key_passphrase="hunter1",
        key_id="success",
        all_principals=True,
    )
    _assert_cert_basic(res, cert_type, rsa_pubkey, ca_pub)


@pytest.mark.skipif(
    HAS_BCRYPT is False, reason="Encrypted keys require the bcrypt library"
)
@pytest.mark.parametrize("cert_type", ["user", "host"])
def test_create_certificate_from_encrypted_privkey_with_encrypted_privkey(
    ssh, ca_key_enc, ca_pub, rsa_privkey_enc, rsa_pubkey, cert_type
):
    res = ssh.create_certificate(
        cert_type=cert_type,
        signing_private_key=ca_key_enc,
        signing_private_key_passphrase="correct horse battery staple",
        private_key=rsa_privkey_enc,
        private_key_passphrase="hunter1",
        key_id="success",
        all_principals=True,
    )
    _assert_cert_basic(res, cert_type, rsa_pubkey, ca_pub)


@pytest.mark.parametrize("cert_type", ["user", "host"])
@pytest.mark.parametrize(
    "algo",
    [
        "rsa",
        "ec",
        pytest.param("ed25519", marks=pytest.mark.skip_on_fips_enabled_platform),
    ],
)
def test_create_certificate_from_pubkey(ssh, ca_key, ca_pub, algo, request, cert_type):
    pubkey = request.getfixturevalue(f"{algo}_pubkey")
    res = ssh.create_certificate(
        cert_type=cert_type,
        signing_private_key=ca_key,
        public_key=pubkey,
        key_id="success",
        all_principals=True,
    )
    _assert_cert_basic(res, cert_type, pubkey, ca_pub)


def test_create_certificate_write_to_path(ssh, ca_key, rsa_pubkey, ca_pub, tmp_path):
    tgt = tmp_path / "cert"
    ssh.create_certificate(
        cert_type="user",
        public_key=rsa_pubkey,
        signing_private_key=ca_key,
        key_id="success",
        all_principals=True,
        path=str(tgt),
    )
    assert tgt.exists()
    _assert_cert_basic(
        tgt.read_bytes(), serialization.SSHCertificateType.USER, rsa_pubkey, ca_pub
    )


def test_create_certificate_write_to_path_overwrite(
    ssh, ca_key, rsa_pubkey, ca_pub, tmp_path
):
    tgt = tmp_path / "cert"
    tgt.write_text("occupied")
    assert tgt.exists()
    ssh.create_certificate(
        cert_type="user",
        overwrite=True,
        public_key=rsa_pubkey,
        signing_private_key=ca_key,
        key_id="success",
        all_principals=True,
        path=str(tgt),
    )
    _assert_cert_basic(
        tgt.read_bytes(), serialization.SSHCertificateType.USER, rsa_pubkey, ca_pub
    )


def test_create_certificate_write_to_path_overwrite_false(
    ssh, ca_key, rsa_pubkey, tmp_path
):
    tgt = tmp_path / "cert"
    tgt.write_text("occupied")
    assert tgt.exists()
    with pytest.raises(
        salt.exceptions.CommandExecutionError,
        match=".*exists and overwrite was set to false.*",
    ):
        ssh.create_certificate(
            "user",
            overwrite=False,
            public_key=rsa_pubkey,
            signing_private_key=ca_key,
            key_id="success",
            all_principals=True,
            path=str(tgt),
        )
    assert tgt.read_text() == "occupied"


def test_create_certificate_raw(ssh, ca_key, ca_pub, rsa_pubkey):
    res = ssh.create_certificate(
        cert_type="user",
        raw=True,
        public_key=rsa_pubkey,
        signing_private_key=ca_key,
        key_id="success",
        all_principals=True,
    )
    assert isinstance(res, bytes)
    _assert_cert_basic(res, serialization.SSHCertificateType.USER, rsa_pubkey, ca_pub)


def test_create_certificate_with_critical_options(ssh, ca_key, ca_pub, rsa_pubkey):
    options = {
        "force-command": "echo hi",
        "no-port-forwarding": True,
        "verify-required": False,
    }
    expected_options = {b"force-command": b"echo hi", b"no-port-forwarding": b""}
    res = ssh.create_certificate(
        cert_type="user",
        signing_private_key=ca_key,
        public_key=rsa_pubkey,
        critical_options=options,
        key_id="success",
        all_principals=True,
    )
    cert = _assert_cert_basic(
        res, serialization.SSHCertificateType.USER, rsa_pubkey, ca_pub
    )
    assert cert.critical_options == expected_options


def test_create_certificate_with_extensions(ssh, ca_key, ca_pub, rsa_pubkey):
    extensions = {
        "permit-X11-forwarding": True,
        "permit-agent-forwarding": True,
        "permit-port-forwarding": True,
        "permit-pty": True,
        "permit-user-rc": False,
    }
    expected_extensions = {
        b"permit-X11-forwarding": b"",
        b"permit-agent-forwarding": b"",
        b"permit-port-forwarding": b"",
        b"permit-pty": b"",
    }
    res = ssh.create_certificate(
        cert_type="user",
        signing_private_key=ca_key,
        public_key=rsa_pubkey,
        extensions=extensions,
        key_id="success",
        all_principals=True,
    )
    cert = _assert_cert_basic(
        res, serialization.SSHCertificateType.USER, rsa_pubkey, ca_pub
    )
    assert cert.extensions == expected_extensions


def test_create_certificate_with_signing_policy_host(ssh, ca_pub, ca_key, rsa_pubkey):
    res = ssh.create_certificate(
        cert_type="user",
        signing_policy="testhostpolicy",
        signing_private_key=ca_key,
        public_key=rsa_pubkey,
        key_id="success",
        valid_principals=["from_call"],
        extensions={"permit-agent-forwarding": True},
        critical_options={"force-command": "rm -rf /"},
    )
    cert = _assert_cert_basic(
        res,
        serialization.SSHCertificateType.HOST,
        rsa_pubkey,
        ca_pub,
        [b"from_host_signing_policy"],
    )
    assert cert.extensions == {}
    assert cert.critical_options == {}


def test_create_certificate_with_signing_policy_user(ssh, ca_pub, ca_key, rsa_pubkey):
    critical_options = {
        "force-command": "rm -rf /",
        "verify-required": True,
        "source-address": "13.37.13.37",
    }
    expected_options = {
        b"force-command": b"echo hi",
        b"no-port-forwarding": b"",
        b"source-address": b"13.37.13.37",
    }
    extensions = {"permit-user-rc": True, "no-touch-required": True}
    expected_extensions = {
        b"permit-X11-forwarding": b"",
        b"permit-agent-forwarding": b"",
        b"permit-port-forwarding": b"",
        b"permit-pty": b"",
        b"no-touch-required": b"",
    }
    res = ssh.create_certificate(
        cert_type="host",
        signing_policy="testuserpolicy",
        signing_private_key=ca_key,
        public_key=rsa_pubkey,
        key_id="success",
        valid_principals=["from_call"],
        critical_options=critical_options,
        extensions=extensions,
    )
    cert = _assert_cert_basic(
        res,
        serialization.SSHCertificateType.USER,
        rsa_pubkey,
        ca_pub,
        [b"from_user_signing_policy"],
    )
    assert cert.critical_options == expected_options
    assert cert.extensions == expected_extensions


def test_create_certificate_with_signing_policy_principals_is_subset(
    ssh, ca_key, ca_pub, rsa_pubkey
):
    """
    Only principals listed in the policy should be allowed
    """
    res = ssh.create_certificate(
        signing_policy="testprincipalspolicy",
        signing_private_key=ca_key,
        public_key=rsa_pubkey,
        valid_principals=["a", "c", "d"],
        key_id="success",
    )
    cert = _get_cert(res)
    assert cert.type is serialization.SSHCertificateType.HOST
    assert cert.key_id == b"success"
    assert _encode_pubkey(cert.public_key()) == rsa_pubkey
    assert _encode_pubkey(cert.signature_key()) == ca_pub
    assert set(cert.valid_principals) == {b"a", b"c"}


def test_create_certificate_not_before_not_after(ssh, ca_pub, ca_key, rsa_pubkey):
    fmt = "%Y-%m-%d %H:%M:%S"
    not_before = "2022-12-21 13:37:10"
    not_after = "2032-12-21 13:37:10"
    res = ssh.create_certificate(
        cert_type="host",
        not_before=not_before,
        not_after=not_after,
        signing_private_key=ca_key,
        public_key=rsa_pubkey,
        key_id="success",
        all_principals=True,
    )
    cert = _assert_cert_basic(
        res, serialization.SSHCertificateType.HOST, rsa_pubkey, ca_pub
    )
    assert cert.valid_after == datetime.datetime.strptime(not_before, fmt).timestamp()
    assert cert.valid_before == datetime.datetime.strptime(not_after, fmt).timestamp()


@pytest.mark.parametrize("sn", [3735928559, "DE:AD:BE:EF", "deadbeef"])
def test_create_certificate_explicit_serial_number(ssh, ca_pub, ca_key, rsa_pubkey, sn):
    res = ssh.create_certificate(
        cert_type="host",
        serial_number=sn,
        signing_private_key=ca_key,
        public_key=rsa_pubkey,
        key_id="success",
        all_principals=True,
    )
    cert = _assert_cert_basic(
        res, serialization.SSHCertificateType.HOST, rsa_pubkey, ca_pub
    )
    assert cert.serial == 3735928559


def test_create_certificate_copypath(ssh, rsa_pubkey, ca_pub, ca_key, tmp_path):
    res = ssh.create_certificate(
        cert_type="host",
        signing_private_key=ca_key,
        public_key=rsa_pubkey,
        key_id="success",
        all_principals=True,
        copypath=str(tmp_path),
    )
    cert = _assert_cert_basic(
        res, serialization.SSHCertificateType.HOST, rsa_pubkey, ca_pub
    )
    target_path = tmp_path / f"{cert.serial:x}.crt"
    assert target_path.exists()
    assert target_path.read_bytes() == cert.public_bytes()


@pytest.mark.slow_test
@pytest.mark.parametrize(
    "algo",
    [
        "rsa",
        "ec",
        pytest.param("ed25519", marks=pytest.mark.skip_on_fips_enabled_platform),
    ],
)
def test_create_private_key(ssh, algo):
    res = ssh.create_private_key(algo=algo)
    priv, pub = res["private_key"], res["public_key"]
    assert priv.startswith("-----BEGIN OPENSSH PRIVATE KEY-----")
    # ensure it can be loaded
    assert ssh.get_public_key(priv) == pub


@pytest.mark.skipif(
    HAS_BCRYPT is False, reason="Encrypted keys require the bcrypt library"
)
@pytest.mark.slow_test
@pytest.mark.parametrize(
    "algo",
    [
        "rsa",
        "ec",
        pytest.param("ed25519", marks=pytest.mark.skip_on_fips_enabled_platform),
    ],
)
def test_create_private_key_with_passphrase(ssh, algo):
    passphrase = "hunter2"
    res = ssh.create_private_key(algo=algo, passphrase=passphrase)
    priv, pub = res["private_key"], res["public_key"]
    assert priv.startswith("-----BEGIN OPENSSH PRIVATE KEY-----")
    # ensure it can be loaded
    assert ssh.get_public_key(priv, passphrase=passphrase) == pub


def test_create_private_key_write_to_path(ssh, tmp_path):
    tgt = tmp_path / "pk"
    pub = tgt.with_suffix(".pub")
    ssh.create_private_key(path=str(tgt))
    assert tgt.exists()
    assert pub.exists()
    assert tgt.read_text().startswith("-----BEGIN OPENSSH PRIVATE KEY-----")
    assert stat.S_IMODE(tgt.stat().st_mode) == 0o0600
    assert ssh.get_public_key(str(tgt)) == pub.read_text()


def test_create_private_key_write_to_path_overwrite(ssh, tmp_path):
    tgt = tmp_path / "pk"
    tgt.write_text("occupied")
    assert tgt.exists()
    ssh.create_private_key(overwrite=True, path=str(tgt))
    assert tgt.read_text().startswith("-----BEGIN OPENSSH PRIVATE KEY-----")
    # ensure it can be loaded
    ssh.get_private_key_size(str(tgt))


def test_create_private_key_write_to_path_not_overwrite(ssh, tmp_path):
    tgt = tmp_path / "pk"
    tgt.write_text("occupied")
    assert tgt.exists()
    with pytest.raises(
        salt.exceptions.CommandExecutionError,
        match=".*exists and overwrite was set to false.*",
    ):
        ssh.create_private_key(overwrite=False, path=str(tgt))
    assert tgt.read_text() == "occupied"


def test_create_private_key_raw(ssh):
    res = ssh.create_private_key(raw=True)
    assert isinstance(res, bytes)
    assert res.startswith(b"-----BEGIN OPENSSH PRIVATE KEY-----")
    # ensure it can be loaded
    ssh.get_private_key_size(res)


@pytest.mark.parametrize(
    "algo,expected",
    [
        ("rsa", 2048),
        ("ec", 256),
        pytest.param("ed25519", None, marks=pytest.mark.skip_on_fips_enabled_platform),
    ],
)
def test_get_private_key_size(ssh, algo, expected, request):
    privkey = request.getfixturevalue(f"{algo}_privkey")
    res = ssh.get_private_key_size(privkey)
    assert res == expected


@pytest.mark.parametrize("source", ["rsa_privkey", "rsa_pubkey", "cert_exts"])
def test_get_public_key(ssh, source, request, rsa_pubkey):
    src = request.getfixturevalue(source)
    res = ssh.get_public_key(src)
    assert res == rsa_pubkey


@pytest.mark.skip_on_fips_enabled_platform
def test_read_certificate(ssh, cert_exts, cert_exts_read):
    res = ssh.read_certificate(cert_exts)
    assert res == cert_exts_read


@pytest.mark.parametrize(
    "priv,pub,expected",
    [
        ("rsa_privkey", "cert_exts", True),
        ("ec_privkey", "ec_pubkey", True),
        pytest.param(
            "ed25519_privkey",
            "ed25519_pubkey",
            True,
            marks=pytest.mark.skip_on_fips_enabled_platform,
        ),
        pytest.param(
            "ed25519_privkey",
            "ec_pubkey",
            False,
            marks=pytest.mark.skip_on_fips_enabled_platform,
        ),
        ("rsa_privkey", "ca_pub", False),
    ],
)
def test_verify_private_key(ssh, priv, pub, expected, request):
    priv = request.getfixturevalue(priv)
    pub = request.getfixturevalue(pub)
    assert ssh.verify_private_key(priv, pub) is expected


@pytest.mark.parametrize(
    "algo",
    [
        "rsa",
        "ec",
        pytest.param("ed25519", marks=pytest.mark.skip_on_fips_enabled_platform),
    ],
)
def test_verify_signature(ssh, algo, request):
    wrong_privkey = request.getfixturevalue(f"{algo}_privkey")
    privkey = ssh.create_private_key(algo=algo)["private_key"]
    privkey_to_sign = ssh.create_private_key()["private_key"]
    cert = ssh.create_certificate(
        cert_type="host",
        all_principals=True,
        private_key=privkey_to_sign,
        signing_private_key=privkey,
    )
    assert ssh.verify_signature(cert, privkey)
    assert not ssh.verify_signature(cert, wrong_privkey)


@pytest.fixture
def fresh_cert(ssh, rsa_privkey, ca_key):
    return ssh.create_certificate(
        cert_type="host",
        private_key=rsa_privkey,
        signing_private_key=ca_key,
        ttl="1d",
        key_id="fresh",
        all_principals=True,
    )


@pytest.mark.parametrize(
    "ttl,expected", [(0, False), (60, False), ("1h", False), ("2d", True)]
)
def test_expires(ssh, fresh_cert, ttl, expected):
    assert ssh.expires(fresh_cert, ttl=ttl) is expected


def _get_cert(cert):
    if isinstance(cert, str):
        cert = cert.encode()
    ret = serialization.load_ssh_public_identity(cert)
    if not isinstance(ret, serialization.SSHCertificate):
        raise ValueError(f"Expected SSHCertificate, got {ret.__class__.__name__}")
    return ret


def _encode_pubkey(public_key):
    return public_key.public_bytes(
        encoding=serialization.Encoding.OpenSSH,
        format=serialization.PublicFormat.OpenSSH,
    ).decode()



"""
integration tests for mac_softwareupdate
"""

import pytest

from salt.exceptions import SaltInvocationError

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.timeout(600, func_only=True),
    pytest.mark.skip_if_binaries_missing("softwareupdate"),
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="module")
def softwareupdate(modules):
    return modules.softwareupdate


@pytest.fixture
def _reset_schedule_enabled(softwareupdate):
    ret = softwareupdate.schedule_enabled()
    try:
        yield
    finally:
        softwareupdate.schedule_enable(ret)


@pytest.fixture
def _reset_catalog(softwareupdate):
    ret = softwareupdate.get_catalog()
    try:
        yield
    finally:
        if ret == "Default":
            softwareupdate.reset_catalog()
        else:
            softwareupdate.set_catalog(ret)


@pytest.fixture
def _reset_ignored(softwareupdate):
    ret = softwareupdate.list_ignored() or ()
    try:
        yield
    finally:
        for item in ret:
            softwareupdate.ignore(item)


def test_list_available(softwareupdate):
    """
    Test softwareupdate.list_available
    """
    # Can't predict what will be returned, so can only test that the return
    # is the correct type, dict
    ret = softwareupdate.list_available()
    assert isinstance(ret, dict)


@pytest.mark.usefixtures("_reset_ignored")
@pytest.mark.skip(reason="Ignore removed from latest OS X.")
def test_ignore(softwareupdate):
    """
    Test softwareupdate.ignore
    Test softwareupdate.list_ignored
    Test softwareupdate.reset_ignored
    """
    # Test reset_ignored
    ret = softwareupdate.reset_ignored()
    assert ret

    ret = softwareupdate.list_ignored()
    assert ret == []

    # Test ignore
    ret = softwareupdate.ignore("spongebob")
    assert ret

    ret = softwareupdate.ignore("squidward")
    assert ret

    # Test list_ignored and verify ignore
    ret = softwareupdate.list_ignored()
    assert "spongebob" in ret

    ret = softwareupdate.list_ignored()
    assert "squidward" in ret


@pytest.mark.usefixtures("_reset_schedule_enabled")
@pytest.mark.skip(reason="Ignore schedule support removed from latest OS X.")
def test_schedule(softwareupdate):
    """
    Test softwareupdate.schedule_enable
    Test softwareupdate.schedule_enabled
    """
    # Test enable
    ret = softwareupdate.schedule_enable(True)
    assert ret

    ret = softwareupdate.schedule_enabled()
    assert ret

    # Test disable in case it was already enabled
    ret = softwareupdate.schedule_enable(False)
    assert not ret

    ret = softwareupdate.schedule_enabled()
    assert not ret


def test_update(softwareupdate):
    """
    Test softwareupdate.update_all
    Test softwareupdate.update
    Test softwareupdate.update_available

    Need to know the names of updates that are available to properly test
    the update functions...
    """
    # There's no way to know what the dictionary will contain, so all we can
    # check is that the return is a dictionary
    ret = softwareupdate.update_all()
    assert isinstance(ret, dict)

    # Test update_available
    ret = softwareupdate.update_available("spongebob")
    assert not ret

    # Test update not available
    with pytest.raises(SaltInvocationError) as exc:
        ret = softwareupdate.update("spongebob")
        assert "Update not available" in str(exc.value)


def test_list_downloads(softwareupdate):
    """
    Test softwareupdate.list_downloads
    """
    ret = softwareupdate.list_downloads()
    assert isinstance(ret, list)


def test_download(softwareupdate):
    """
    Test softwareupdate.download

    Need to know the names of updates that are available to properly test
    the download function
    """
    # Test update not available
    with pytest.raises(SaltInvocationError) as exc:
        softwareupdate.download("spongebob")
        assert "Update not available" in str(exc.value)


def test_download_all(softwareupdate):
    """
    Test softwareupdate.download_all
    """
    ret = softwareupdate.download_all()
    assert isinstance(ret, list)


@pytest.mark.usefixtures("_reset_catalog")
@pytest.mark.skip(reason="Ignore catalog support removed from latest OS X.")
def test_get_set_reset_catalog(softwareupdate):
    """
    Test softwareupdate.download_all
    """
    # Reset the catalog
    ret = softwareupdate.reset_catalog()
    assert ret

    ret = softwareupdate.get_catalog()
    assert ret == "Default"

    # Test setting and getting the catalog
    ret = softwareupdate.set_catalog("spongebob")
    assert ret

    ret = softwareupdate.get_catalog()
    assert ret == "spongebob"

    # Test reset the catalog
    ret = softwareupdate.reset_catalog()
    assert ret

    assert softwareupdate.get_catalog()
    assert ret == "Default"



import tempfile
from datetime import datetime

import pytest

import salt.modules.file
import salt.modules.macdefaults as macdefaults
from tests.support.mock import ANY, MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {macdefaults: {}}


@pytest.fixture
def PLIST_OUTPUT():
    return """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
	<key>AdjustWindowForFontSizeChange</key>
	<false/>
	<key>AllowClipboardAccess</key>
	<true/>
	<key>AppleAntiAliasingThreshold</key>
	<integer>1024</integer>
	<key>Default Bookmark Guid</key>
	<string>C7EED71F-6B5F-4822-B735-D20CAE8AD57D</string>
	<key>NSNavPanelExpandedSizeForOpenMode</key>
	<string>{800, 448}</string>
	<key>NSSplitView Subview Frames NSColorPanelSplitView</key>
	<array>
		<string>0.000000, 0.000000, 224.000000, 222.000000, NO, NO</string>
		<string>0.000000, 223.000000, 224.000000, 48.000000, NO, NO</string>
	</array>
	<key>NSToolbar Configuration com.apple.NSColorPanel</key>
	<dict>
		<key>TB Is Shown</key>
		<integer>1</integer>
	</dict>
	<key>NeverWarnAboutShortLivedSessions_C7EED71F-6B5F-4822-B735-D20CAE8AD57D_selection</key>
	<integer>0</integer>
	<key>NoSyncBFPRecents</key>
	<array>
		<string>MonoLisa Variable</string>
		<string>Menlo</string>
		<string>MonoLisa</string>
	</array>
	<key>NoSyncFrame_SharedPreferences</key>
	<dict>
		<key>screenFrame</key>
		<string>{{0, 0}, {1920, 1200}}</string>
		<key>topLeft</key>
		<string>{668, 1004}</string>
	</dict>
	<key>NoSyncNextAnnoyanceTime</key>
	<real>699041775.9120981</real>
	<key>NoSyncTipOfTheDayEligibilityBeganTime</key>
	<real>697356008.723986</real>
	<key>PointerActions</key>
	<dict>
		<key>Button,1,1,,</key>
		<dict>
			<key>Action</key>
			<string>kContextMenuPointerAction</string>
		</dict>
		<key>Button,2,1,,</key>
		<dict>
			<key>Action</key>
			<string>kPasteFromSelectionPointerAction</string>
		</dict>
	</dict>
	<key>SULastCheckTime</key>
	<date>2024-06-23T09:33:44Z</date>
</dict>
</plist>
"""


def test_run_defaults_cmd():
    """
    Test calling _run_defaults_cmd
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": "Server", "stderr": ""})
    with patch.dict(macdefaults.__salt__, {"cmd.run_all": mock}):
        result = macdefaults._run_defaults_cmd(
            'read "com.apple.CrashReporter" "DialogType"'
        )
        mock.assert_called_once_with(
            'defaults read "com.apple.CrashReporter" "DialogType"', runas=None
        )
        assert result == {"retcode": 0, "stdout": "Server", "stderr": ""}


def test_run_defaults_cmd_with_user():
    """
    Test caling _run_defaults_cmd
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": "Server", "stderr": ""})
    with patch.dict(macdefaults.__salt__, {"cmd.run_all": mock}):
        result = macdefaults._run_defaults_cmd(
            'read "com.apple.CrashReporter" "DialogType"', runas="frank"
        )
        mock.assert_called_once_with(
            'defaults read "com.apple.CrashReporter" "DialogType"', runas="frank"
        )
        assert result == {"retcode": 0, "stdout": "Server", "stderr": ""}


def test_load_plist(PLIST_OUTPUT):
    """
    Test loading a plist
    """
    expected_result = {"Crash": "Server"}
    run_defaults_cmd_mock = MagicMock(
        return_value={"retcode": 0, "stdout": PLIST_OUTPUT, "stderr": ""}
    )

    with patch(
        "salt.modules.macdefaults._run_defaults_cmd", run_defaults_cmd_mock
    ), patch("plistlib.loads", return_value=expected_result) as plist_mock:
        result = macdefaults._load_plist("com.googlecode.iterm2")
        run_defaults_cmd_mock.assert_called_once_with(
            'export "com.googlecode.iterm2" -', runas=None
        )
        plist_mock.assert_called_once_with(PLIST_OUTPUT.encode())
        assert result == expected_result


def test_load_plist_no_domain():
    """
    Test loading a plist with a non-existent domain
    """
    empty_plist = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict/>
</plist>"""

    run_defaults_cmd_mock = MagicMock(
        return_value={"retcode": 0, "stdout": empty_plist, "stderr": ""}
    )

    with patch("salt.modules.macdefaults._run_defaults_cmd", run_defaults_cmd_mock):
        result = macdefaults._load_plist("com.googlecode.iterm2", user="cdalvaro")
        run_defaults_cmd_mock.assert_called_once_with(
            'export "com.googlecode.iterm2" -', runas="cdalvaro"
        )
        assert result is None


def test_save_plist_no_user():
    """
    Test saving a plist
    """
    new_plist = {"Crash": "Server"}
    expected_result = {"retcode": 0, "stdout": "", "stderr": ""}

    chown_mock = MagicMock()

    tempdir = tempfile.TemporaryDirectory()
    tempdir_name = tempdir.name

    tempfile_mock = MagicMock()
    tempfile_mock.__enter__.return_value = tempdir_name

    domain = "com.googlecode.iterm2"
    tmp_file_name = salt.modules.file.join(tempdir_name, f"{domain}.plist")

    with patch(
        "salt.modules.macdefaults._run_defaults_cmd", return_value=expected_result
    ) as run_defaults_cmd_mock, patch(
        "tempfile.TemporaryDirectory", return_value=tempfile_mock
    ), patch(
        "plistlib.dump"
    ) as plist_mock, patch.dict(
        macdefaults.__salt__,
        {"file.chown": chown_mock, "file.join": salt.modules.file.join},
    ):
        result = macdefaults._save_plist("com.googlecode.iterm2", new_plist)
        assert result == expected_result

        plist_mock.assert_called_once_with(new_plist, ANY)  # ANY for filehandler
        run_defaults_cmd_mock.assert_called_once_with(
            f'import "{domain}" "{tmp_file_name}"',
            runas=None,
        )
        chown_mock.assert_not_called()


def test_save_plist_with_user():
    """
    Test saving a plist
    """
    new_plist = {"Crash": "Server"}
    expected_result = {"retcode": 0, "stdout": "", "stderr": ""}

    chown_mock = MagicMock()

    tempdir = tempfile.TemporaryDirectory()
    tempdir_name = tempdir.name

    tempfile_mock = MagicMock()
    tempfile_mock.__enter__.return_value = tempdir_name

    user = "cdalvaro"
    domain = "com.googlecode.iterm2"
    tmp_file_name = salt.modules.file.join(tempdir_name, f"{domain}.plist")

    chown_expected_calls = [
        call(tempdir_name, user, None),
        call(tmp_file_name, user, None),
    ]

    with patch(
        "salt.modules.macdefaults._run_defaults_cmd", return_value=expected_result
    ) as run_defaults_cmd_mock, patch(
        "tempfile.TemporaryDirectory", return_value=tempfile_mock
    ), patch(
        "plistlib.dump"
    ) as plist_mock, patch.dict(
        macdefaults.__salt__,
        {"file.chown": chown_mock, "file.join": salt.modules.file.join},
    ):
        result = macdefaults._save_plist("com.googlecode.iterm2", new_plist, user=user)
        assert result == expected_result

        plist_mock.assert_called_once_with(new_plist, ANY)  # ANY for filehandler
        run_defaults_cmd_mock.assert_called_once_with(
            f'import "{domain}" "{tmp_file_name}"',
            runas=user,
        )
        chown_mock.assert_has_calls(chown_expected_calls)


def test_cast_value_to_bool():
    assert macdefaults.cast_value_to_vtype("TRUE", "bool") is True
    assert macdefaults.cast_value_to_vtype("YES", "bool") is True
    assert macdefaults.cast_value_to_vtype("1", "bool") is True
    assert macdefaults.cast_value_to_vtype(1, "bool") is True

    assert macdefaults.cast_value_to_vtype("FALSE", "boolean") is False
    assert macdefaults.cast_value_to_vtype("NO", "boolean") is False
    assert macdefaults.cast_value_to_vtype("0", "boolean") is False
    assert macdefaults.cast_value_to_vtype(0, "boolean") is False

    with pytest.raises(ValueError):
        macdefaults.cast_value_to_vtype("foo", "bool")

    with pytest.raises(ValueError):
        macdefaults.cast_value_to_vtype(1.1, "bool")


def test_cast_value_to_string():
    assert macdefaults.cast_value_to_vtype(124.120, "string") == "124.12"
    assert macdefaults.cast_value_to_vtype(True, "string") == "YES"
    assert macdefaults.cast_value_to_vtype(False, "string") == "NO"
    assert macdefaults.cast_value_to_vtype(124120, "string") == "124120"

    expected_date = "2024-06-26T16:45:05Z"
    test_date = datetime.strptime(expected_date, "%Y-%m-%dT%H:%M:%SZ")
    assert macdefaults.cast_value_to_vtype(test_date, "string") == expected_date

    expected_data = "foo"
    test_data = expected_data.encode()
    assert macdefaults.cast_value_to_vtype(test_data, "string") == expected_data


def test_cast_value_to_int():
    assert macdefaults.cast_value_to_vtype("124", "int") == 124
    assert macdefaults.cast_value_to_vtype(124.0, "int") == 124
    assert macdefaults.cast_value_to_vtype(True, "integer") == 1
    assert macdefaults.cast_value_to_vtype(False, "integer") == 0

    with pytest.raises(ValueError):
        macdefaults.cast_value_to_vtype("foo", "int")


def test_cast_value_to_float():
    assert macdefaults.cast_value_to_vtype("124.12", "float") == 124.12
    assert macdefaults.cast_value_to_vtype(124, "float") == 124.0
    assert macdefaults.cast_value_to_vtype(True, "float") == 1.0
    assert macdefaults.cast_value_to_vtype(False, "float") == 0.0

    with pytest.raises(ValueError):
        macdefaults.cast_value_to_vtype("foo", "float")


def test_cast_value_to_date():
    expected_date = datetime(2024, 6, 26, 16, 45, 5)

    # Date -> Date
    assert macdefaults.cast_value_to_vtype(expected_date, "date") == expected_date

    # String -> Date
    test_date = datetime.strftime(expected_date, "%Y-%m-%dT%H:%M:%SZ")
    assert macdefaults.cast_value_to_vtype(test_date, "date") == expected_date

    # Timestamp -> Date
    test_timestamp = expected_date.timestamp()
    assert macdefaults.cast_value_to_vtype(test_timestamp, "date") == expected_date

    with pytest.raises(ValueError):
        macdefaults.cast_value_to_vtype("foo", "date")


def test_cast_value_to_data():
    expected_data = b"foo"

    # String -> Data
    test_data = expected_data.decode()
    assert macdefaults.cast_value_to_vtype(test_data, "data") == expected_data

    # Data -> Data
    assert macdefaults.cast_value_to_vtype(expected_data, "data") == expected_data

    with pytest.raises(ValueError):
        macdefaults.cast_value_to_vtype(123, "data")


def test_write_default():
    """
    Test writing a default setting
    """
    with patch("salt.modules.macdefaults._load_plist", return_value={}), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "DialogType", "Server")
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"DialogType": "Server"},
            user=None,
        )


def test_write_default_with_user():
    """
    Test writing a default setting with a specific user
    """
    with patch("salt.modules.macdefaults._load_plist", return_value={}), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write(
            "com.apple.CrashReporter", "DialogType", "Server", user="cdalvaro"
        )
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"DialogType": "Server"},
            user="cdalvaro",
        )


def test_write_default_true_boolean():
    """
    Test writing a default True boolean setting
    """
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": False}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", True)
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": True},
            user=None,
        )

    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": False}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", "YES", vtype="boolean")
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": True},
            user=None,
        )


def test_write_default_false_bool():
    """
    Test writing a default False boolean setting
    """
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": True}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", False)
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": False},
            user=None,
        )

    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": True}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", "NO", vtype="bool")
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": False},
            user=None,
        )


def test_write_default_int():
    """
    Test writing a default int setting
    """
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": 0}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", 1)
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": 1},
            user=None,
        )

    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": 3}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", "3", vtype="int")
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": 3},
            user=None,
        )

    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": 0}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", "15", vtype="integer")
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": 15},
            user=None,
        )


def test_write_default_float():
    """
    Test writing a default float setting
    """
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": 0}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", 1.234)
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": 1.234},
            user=None,
        )

    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": 0}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", "14.350", vtype="float")
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": 14.35},
            user=None,
        )


def test_write_default_array():
    """
    Test writing a default array setting
    """

    # Key does not exist
    with patch("salt.modules.macdefaults._load_plist", return_value={}), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", [0.7, 0.9, 1.0])
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": [0.7, 0.9, 1.0]},
            user=None,
        )

    # Key already exists with different values
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": [0.1, 0.2, 0.4]}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", [0.7, 0.9, 1.0])
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": [0.7, 0.9, 1.0]},
            user=None,
        )

    # Array already exists and the new value (float) is appended
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": [0.1, 0.2, 0.4]}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write(
            "com.apple.CrashReporter", "Crash", "0.5", vtype="float", array_add=True
        )
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": [0.1, 0.2, 0.4, 0.5]},
            user=None,
        )

    # Array already exists and the new value (array) is appended (using array_add=True)
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": [0.1, 0.2, 0.4]}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write(
            "com.apple.CrashReporter", "Crash", [2.0, 4.0], array_add=True
        )
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": [0.1, 0.2, 0.4, 2.0, 4.0]},
            user=None,
        )

    # Array already exists and the new value (array) is appended (using vtype="array-add")
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": [0.1, 0.2, 0.4]}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write(
            "com.apple.CrashReporter", "Crash", [2.0, 4.0], vtype="array-add"
        )
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": [0.1, 0.2, 0.4, 2.0, 4.0]},
            user=None,
        )

    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": [0.1, 0.2, 0.4]}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock, pytest.raises(
        ValueError
    ) as excinfo:
        macdefaults.write("com.apple.CrashReporter", "Crash", "0.5", vtype="array")
        save_plist_mock.assert_not_called()
        excinfo.match("Invalid value for array")


def test_write_default_dictionary():
    """
    Test writing a default dictionary setting
    """

    # Key does not exist
    with patch("salt.modules.macdefaults._load_plist", return_value={}), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", {"foo:": "bar", "baz": 0})
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": {"foo:": "bar", "baz": 0}},
            user=None,
        )

    # Key already exists with different values
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": {"foo": "bar"}}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write("com.apple.CrashReporter", "Crash", {"baz": 1})
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": {"baz": 1}},
            user=None,
        )

    # Dictionary already exists and the new value is merged (using dict_merge=True)
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": {"foo": "bar"}}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write(
            "com.apple.CrashReporter", "Crash", {"baz": 10}, dict_merge=True
        )
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": {"foo": "bar", "baz": 10}},
            user=None,
        )

    # Dictionary already exists and the new value is merged (using vtype="dict-add")
    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": {"foo": "bar"}}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock:
        macdefaults.write(
            "com.apple.CrashReporter", "Crash", {"baz": 10}, vtype="dict-add"
        )
        save_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            {"Crash": {"foo": "bar", "baz": 10}},
            user=None,
        )

    with patch(
        "salt.modules.macdefaults._load_plist", return_value={"Crash": {}}
    ), patch(
        "salt.modules.macdefaults._save_plist", return_value={"retcode": 0}
    ) as save_plist_mock, pytest.raises(
        ValueError
    ) as excinfo:
        macdefaults.write("com.apple.CrashReporter", "Crash", "0.5", vtype="dict")
        save_plist_mock.assert_not_called()
        excinfo.match("Invalid value for dictionary")


def test_read_default_string(PLIST_OUTPUT):
    """
    Test reading a default string setting
    """

    def custom_run_defaults_cmd(action, runas):
        return {"retcode": 0, "stdout": PLIST_OUTPUT}

    mock = MagicMock(side_effect=custom_run_defaults_cmd)
    with patch("salt.modules.macdefaults._run_defaults_cmd", mock):
        result = macdefaults.read("com.googlecode.iterm2", "Default Bookmark Guid")
        mock.assert_called_once_with('export "com.googlecode.iterm2" -', runas=None)
        assert result == "C7EED71F-6B5F-4822-B735-D20CAE8AD57D"


def test_read_default_string_with_user(PLIST_OUTPUT):
    """
    Test reading a default string setting as a specific user
    """

    def custom_run_defaults_cmd(action, runas):
        return {"retcode": 0, "stdout": PLIST_OUTPUT}

    mock = MagicMock(side_effect=custom_run_defaults_cmd)
    with patch("salt.modules.macdefaults._run_defaults_cmd", mock):
        result = macdefaults.read(
            "com.googlecode.iterm2", "NSNavPanelExpandedSizeForOpenMode", user="frank"
        )
        mock.assert_called_once_with('export "com.googlecode.iterm2" -', runas="frank")
        assert result == "{800, 448}"


def test_read_default_integer(PLIST_OUTPUT):
    """
    Test reading a default integer setting
    """

    def custom_run_defaults_cmd(action, runas):
        return {"retcode": 0, "stdout": PLIST_OUTPUT}

    mock = MagicMock(side_effect=custom_run_defaults_cmd)
    with patch("salt.modules.macdefaults._run_defaults_cmd", mock):
        result = macdefaults.read("com.googlecode.iterm2", "AppleAntiAliasingThreshold")
        mock.assert_called_once_with('export "com.googlecode.iterm2" -', runas=None)
        assert result == 1024


def test_read_default_float(PLIST_OUTPUT):
    """
    Test reading a default float setting
    """

    def custom_run_defaults_cmd(action, runas):
        return {"retcode": 0, "stdout": PLIST_OUTPUT}

    mock = MagicMock(side_effect=custom_run_defaults_cmd)
    with patch("salt.modules.macdefaults._run_defaults_cmd", mock):
        result = macdefaults.read("com.googlecode.iterm2", "NoSyncNextAnnoyanceTime")
        mock.assert_called_once_with('export "com.googlecode.iterm2" -', runas=None)
        assert result == 699041775.9120981


def test_read_default_array(PLIST_OUTPUT):
    """
    Test reading a default array setting
    """

    def custom_run_defaults_cmd(action, runas):
        return {"retcode": 0, "stdout": PLIST_OUTPUT}

    mock = MagicMock(side_effect=custom_run_defaults_cmd)
    with patch("salt.modules.macdefaults._run_defaults_cmd", mock):
        result = macdefaults.read("com.googlecode.iterm2", "NoSyncBFPRecents")
        mock.assert_called_once_with('export "com.googlecode.iterm2" -', runas=None)
        assert result == ["MonoLisa Variable", "Menlo", "MonoLisa"]


def test_read_default_dictionary(PLIST_OUTPUT):
    """
    Test reading a default dictionary setting
    """

    def custom_run_defaults_cmd(action, runas):
        return {"retcode": 0, "stdout": PLIST_OUTPUT}

    mock = MagicMock(side_effect=custom_run_defaults_cmd)
    with patch("salt.modules.macdefaults._run_defaults_cmd", mock):
        result = macdefaults.read(
            "com.googlecode.iterm2", "NoSyncFrame_SharedPreferences"
        )
        mock.assert_called_once_with('export "com.googlecode.iterm2" -', runas=None)
        assert result == {
            "screenFrame": "{{0, 0}, {1920, 1200}}",
            "topLeft": "{668, 1004}",
        }


def test_read_default_dictionary_nested_key(PLIST_OUTPUT):
    """
    Test reading a default dictionary setting with a nested key
    """

    def custom_run_defaults_cmd(action, runas):
        return {"retcode": 0, "stdout": PLIST_OUTPUT}

    mock = MagicMock(side_effect=custom_run_defaults_cmd)
    with patch("salt.modules.macdefaults._run_defaults_cmd", mock):
        result = macdefaults.read(
            "com.googlecode.iterm2",
            "PointerActions.Button,1,1,,.Action",
            key_separator=".",
        )
        mock.assert_called_once_with('export "com.googlecode.iterm2" -', runas=None)
        assert result == "kContextMenuPointerAction"


def test_read_default_dictionary_nested_key_with_array_indexes(PLIST_OUTPUT):
    """
    Test reading a default dictionary setting with a nested key and array indexes
    """

    def custom_run_defaults_cmd(action, runas):
        return {"retcode": 0, "stdout": PLIST_OUTPUT}

    mock = MagicMock(side_effect=custom_run_defaults_cmd)
    with patch("salt.modules.macdefaults._run_defaults_cmd", mock):
        # First index (exists)
        result = macdefaults.read(
            "com.googlecode.iterm2",
            "NSSplitView Subview Frames NSColorPanelSplitView.0",
            key_separator=".",
        )
        mock.assert_called_once_with('export "com.googlecode.iterm2" -', runas=None)
        assert result == "0.000000, 0.000000, 224.000000, 222.000000, NO, NO"

        # Second index (exists)
        result = macdefaults.read(
            "com.googlecode.iterm2",
            "NSSplitView Subview Frames NSColorPanelSplitView.1",
            key_separator=".",
        )
        assert result == "0.000000, 223.000000, 224.000000, 48.000000, NO, NO"

        # Third index (does not exist)
        result = macdefaults.read(
            "com.googlecode.iterm2",
            "NSSplitView Subview Frames NSColorPanelSplitView.2",
            key_separator=".",
        )
        assert result is None


def test_read_default_date(PLIST_OUTPUT):
    """
    Test reading a default date setting
    """

    def custom_run_defaults_cmd(action, runas):
        return {"retcode": 0, "stdout": PLIST_OUTPUT}

    mock = MagicMock(side_effect=custom_run_defaults_cmd)
    with patch("salt.modules.macdefaults._run_defaults_cmd", mock):
        result = macdefaults.read("com.googlecode.iterm2", "SULastCheckTime")
        mock.assert_called_once_with('export "com.googlecode.iterm2" -', runas=None)
        assert result == datetime.strptime("2024-06-23T09:33:44Z", "%Y-%m-%dT%H:%M:%SZ")


def test_delete_default():
    """
    Test delete a default setting
    """
    original_plist = {
        "Crash": "bar",
        "foo": 0,
    }

    updated_plist = {
        "foo": 0,
    }

    result = {"retcode": 0, "stdout": "Removed key", "stderr": ""}

    load_plist_mock = MagicMock(return_value=original_plist)
    export_plist_mock = MagicMock(return_value=result)

    with patch("salt.modules.macdefaults._load_plist", load_plist_mock), patch(
        "salt.modules.macdefaults._save_plist", export_plist_mock
    ):
        assert result == macdefaults.delete("com.apple.CrashReporter", "Crash")
        load_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            user=None,
        )
        export_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            updated_plist,
            user=None,
        )


def test_delete_default_with_user():
    """
    Test delete a default setting as a specific user
    """
    original_plist = {
        "Crash": {
            "foo": "bar",
            "baz": 0,
        },
        "Crash.baz": 0,
    }

    updated_plist = {
        "Crash": {
            "foo": "bar",
            "baz": 0,
        },
    }

    result = {"retcode": 0, "stdout": "Removed key", "stderr": ""}

    load_plist_mock = MagicMock(return_value=original_plist)
    export_plist_mock = MagicMock(return_value=result)

    with patch("salt.modules.macdefaults._load_plist", load_plist_mock), patch(
        "salt.modules.macdefaults._save_plist", export_plist_mock
    ):
        macdefaults.delete("com.apple.CrashReporter", "Crash.baz", user="frank")
        load_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            user="frank",
        )
        export_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            updated_plist,
            user="frank",
        )


def test_delete_default_with_nested_key():
    """
    Test delete a default setting with a nested key
    """
    original_plist = {
        "Crash": {
            "foo": "bar",
            "baz": 0,
        }
    }

    updated_plist = {
        "Crash": {
            "foo": "bar",
        }
    }

    result = {"retcode": 0, "stdout": "Removed key", "stderr": ""}

    load_plist_mock = MagicMock(return_value=original_plist)
    export_plist_mock = MagicMock(return_value=result)

    with patch("salt.modules.macdefaults._load_plist", load_plist_mock), patch(
        "salt.modules.macdefaults._save_plist", export_plist_mock
    ):
        assert result == macdefaults.delete(
            "com.apple.CrashReporter",
            "Crash.baz",
            key_separator=".",
        )
        load_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            user=None,
        )
        export_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            updated_plist,
            user=None,
        )


def test_delete_default_dictionary_nested_key_with_array_indexes():
    """
    Test delete a default dictionary setting with a nested key and array indexes
    """
    original_plist = {
        "Crash": {
            "foo": "bar",
            "baz": [
                {"internalKey1": 1, "internalKey2": "a"},
                {"internalKey1": 2, "internalKey2": "b"},
                {"internalKey1": 3, "internalKey2": "c"},
            ],
        }
    }

    updated_plist = {
        "Crash": {
            "foo": "bar",
            "baz": [
                {"internalKey1": 1, "internalKey2": "a"},
                {"internalKey2": "b"},
                {"internalKey1": 3, "internalKey2": "c"},
            ],
        }
    }

    result = {"retcode": 0, "stdout": "Removed key", "stderr": ""}

    load_plist_mock = MagicMock(return_value=original_plist)
    export_plist_mock = MagicMock(return_value=result)

    with patch("salt.modules.macdefaults._load_plist", load_plist_mock), patch(
        "salt.modules.macdefaults._save_plist", export_plist_mock
    ):
        assert result == macdefaults.delete(
            "com.apple.CrashReporter",
            "Crash.baz.1.internalKey1",
            key_separator=".",
        )
        load_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            user=None,
        )
        export_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            updated_plist,
            user=None,
        )


def test_delete_default_dictionary_nested_key_with_array_index_as_last_key():
    """
    Test delete a default dictionary setting with a nested key and array indexes
    and the last element of the key is an array index
    """
    original_plist = {
        "Crash": {
            "foo": "bar",
            "baz": [
                {"internalKey1": 1, "internalKey2": "a"},
                {"internalKey1": 2, "internalKey2": "b"},
                {"internalKey1": 3, "internalKey2": "c"},
            ],
        }
    }

    updated_plist = {
        "Crash": {
            "foo": "bar",
            "baz": [
                {"internalKey1": 1, "internalKey2": "a"},
                {"internalKey1": 3, "internalKey2": "c"},
            ],
        }
    }

    result = {"retcode": 0, "stdout": "Removed key", "stderr": ""}

    load_plist_mock = MagicMock(return_value=original_plist)
    export_plist_mock = MagicMock(return_value=result)

    with patch("salt.modules.macdefaults._load_plist", load_plist_mock), patch(
        "salt.modules.macdefaults._save_plist", export_plist_mock
    ):
        assert result == macdefaults.delete(
            "com.apple.CrashReporter",
            "Crash.baz.1",
            key_separator=".",
        )
        load_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            user=None,
        )
        export_plist_mock.assert_called_once_with(
            "com.apple.CrashReporter",
            updated_plist,
            user=None,
        )



"""
Use this script to dump the event data out to the terminal. It needs to know
what the sock_dir is.

This script is a generic tool to test event output
"""

import optparse  # pylint: disable=deprecated-module
import os
import pprint
import time

import salt.utils.event


def parse():
    """
    Parse the script command line inputs
    """
    parser = optparse.OptionParser()

    parser.add_option(
        "-s",
        "--sock-dir",
        dest="sock_dir",
        default="/var/run/salt",
        help=(
            "Statically define the directory holding the salt unix "
            "sockets for communication"
        ),
    )
    parser.add_option(
        "-n",
        "--node",
        dest="node",
        default="master",
        help=(
            "State if this listener will attach to a master or a "
            'minion daemon, pass "master" or "minion"'
        ),
    )
    parser.add_option(
        "-f",
        "--func_count",
        default="",
        help=(
            "Return a count of the number of minions which have "
            "replied to a job with a given func."
        ),
    )
    parser.add_option(
        "-i",
        "--id",
        default="",
        help="If connecting to a live master or minion, pass in the id",
    )
    parser.add_option(
        "-t",
        "--transport",
        default="zeromq",
        help="Transport to use. (Default: 'zeromq'",
    )

    options, args = parser.parse_args()

    opts = {}

    for k, v in options.__dict__.items():
        if v is not None:
            opts[k] = v

    opts["sock_dir"] = os.path.join(opts["sock_dir"], opts["node"])

    if "minion" in options.node:
        if args:
            opts["id"] = args[0]
            return opts
        if options.id:
            opts["id"] = options.id
        else:
            opts["id"] = options.node

    return opts


def check_access_and_print_warning(sock_dir):
    """
    Check if this user is able to access the socket
    directory and print a warning if not
    """
    if (
        os.access(sock_dir, os.R_OK)
        and os.access(sock_dir, os.W_OK)
        and os.access(sock_dir, os.X_OK)
    ):
        return
    else:
        print(
            "WARNING: Events will not be reported (not able to access {})".format(
                sock_dir
            )
        )


def listen(opts):
    """
    Attach to the pub socket and grab messages
    """
    event = salt.utils.event.get_event(
        opts["node"],
        sock_dir=opts["sock_dir"],
        opts=opts,
        listen=True,
    )
    check_access_and_print_warning(opts["sock_dir"])
    print(event.puburi)
    jid_counter = 0
    found_minions = []
    while True:
        ret = event.get_event(full=True)
        if ret is None:
            continue
        if opts["func_count"]:
            data = ret.get("data", False)
            if data:
                if "id" in data.keys() and data.get("id", False) not in found_minions:
                    if data["fun"] == opts["func_count"]:
                        jid_counter += 1
                        found_minions.append(data["id"])
                        print(
                            "Reply received from [{}]. Total replies now: [{}].".format(
                                ret["data"]["id"], jid_counter
                            )
                        )
                    continue
        else:
            print(f"Event fired at {time.asctime()}")
            print("*" * 25)
            print("Tag: {}".format(ret["tag"]))
            print("Data:")
            pprint.pprint(ret["data"])


if __name__ == "__main__":
    opts = parse()
    listen(opts)



"""
Execute batch runs.

The sync CLI driver (``Batch``) is a thin shell around the shared
state machine in :mod:`salt.utils.batch_state` and the CLI output
adapter in :mod:`salt.utils.batch_output`.  The state machine owns
slot accounting, failhard, timeout, and ``batch_wait`` — the driver
only manages the iterator plumbing that turns master-side
``cmd_iter_no_block`` polling into ``new_returns`` inputs and
``timed_out`` signals.

The observable behavior (yield shape, stdout formatting, JID reuse,
failhard early-exit, ``batch_wait`` dispatch delay) is preserved
byte-for-byte against the pre-refactor implementation.  The async
``BatchManager`` uses the same state machine, so both drivers are
guaranteed to produce the same sequence of minion transitions given
the same inputs.
"""

import copy
import logging
import math
import time

import salt.client
import salt.exceptions
import salt.output
import salt.utils.batch_output
import salt.utils.batch_state
import salt.utils.jid
import salt.utils.stringutils

log = logging.getLogger(__name__)


class Batch:
    """
    Manage the execution of batch runs.
    """

    def __init__(self, opts, eauth=None, quiet=False, _parser=None):
        """
        :param dict opts: A config options dictionary.

        :param dict eauth: An eauth config to use.

                           The default is an empty dict.

        :param bool quiet: Suppress printing to stdout

                           The default is False.
        """
        self.opts = opts
        self.eauth = eauth if eauth else {}
        self.pub_kwargs = eauth if eauth else {}
        self.quiet = quiet
        self.options = _parser
        # Passing listen True to local client will prevent it from purging
        # cached events while iterating over the batches.
        self.local = salt.client.get_local_client(opts["conf_file"], listen=True)

    def gather_minions(self):
        """
        Return a list of minions to use for the batch run.
        """
        args = [
            self.opts["tgt"],
            "test.ping",
            [],
            self.opts["timeout"],
        ]

        selected_target_option = self.opts.get("selected_target_option", None)
        if selected_target_option is not None:
            args.append(selected_target_option)
        else:
            args.append(self.opts.get("tgt_type", "glob"))

        self.pub_kwargs["yield_pub_data"] = True
        ping_gen = self.local.cmd_iter(
            *args,
            gather_job_timeout=self.opts["gather_job_timeout"],
            **self.pub_kwargs,
        )

        fret = set()
        nret = set()
        for ret in ping_gen:
            if ("minions" and "jid") in ret:
                for minion in ret["minions"]:
                    nret.add(minion)
                continue
            else:
                try:
                    m = next(iter(ret.keys()))
                except StopIteration:
                    if not self.quiet:
                        salt.utils.stringutils.print_cli(
                            "No minions matched the target."
                        )
                    break
                if m is not None:
                    if "failed" in ret[m] and ret[m]["failed"] is True:
                        log.debug(
                            "minion '%s' failed test.ping - will be returned as a down minion",
                            m,
                        )
                    else:
                        fret.add(m)

        return (list(fret), ping_gen, nret.difference(fret))

    def get_bnum(self):
        """
        Return the active number of minions to maintain.

        Preserves the legacy return values (``None`` for invalid
        input, ``0`` for an empty minion list with a percentage spec,
        ``0`` for ``batch=0``) for backward compatibility with any
        callers that rely on them.  The shared state machine uses the
        hardened :func:`salt.utils.batch_state.get_batch_size` which
        always returns at least 1.
        """

        def partition(x):
            return float(x) / 100.0 * len(self.minions)

        try:
            if isinstance(self.opts["batch"], str) and "%" in self.opts["batch"]:
                res = partition(float(self.opts["batch"].strip("%")))
                if res < 1:
                    return int(math.ceil(res))
                return int(res)
            return int(self.opts["batch"])
        except ValueError:
            if not self.quiet:
                salt.utils.stringutils.print_cli(
                    "Invalid batch data sent: {}\nData must be in the "
                    "form of %10, 10% or 3".format(self.opts["batch"])
                )

    def run(self):
        """
        Execute the batch run.

        Generator.  For each minion return, yields
        ``({minion_id: ret_data}, retcode)`` (or the raw event envelope
        when ``raw=True``).  Minion returns carrying ``failed: True``
        (and ``failhard`` halt events) are recorded internally but not
        yielded, preserving the pre-Phase-2 yield shape.
        """
        self.minions, self.ping_gen, self.down_minions = self.gather_minions()

        if not self.minions:
            return

        batch_jid = salt.utils.jid.gen_jid(self.opts)
        state = salt.utils.batch_state.create_batch_state(
            self.opts, self.minions, batch_jid, driver="cli"
        )

        salt.utils.batch_state.write_batch_state(
            batch_jid, state, self.opts, best_effort=True
        )
        salt.utils.batch_state.add_to_active_index(
            batch_jid, self.opts, best_effort=True
        )

        output = salt.utils.batch_output.CLIOutput(self.opts, quiet=self.quiet)
        for down_minion in self.down_minions:
            output.on_minion_down(down_minion)

        if self.options:
            show_jid = self.options.show_jid
            show_verbose = self.options.verbose
        else:
            show_jid = False
            show_verbose = False

        return_value = self.opts.get("return", self.opts.get("ret", ""))
        raw_mode = bool(self.opts.get("raw"))

        iters = []
        minion_tracker = {}  # iter -> {"minions": [...], "active": bool}
        # We retain the raw return objects (event envelopes in raw mode,
        # minion-return dicts otherwise) separately from the normalized
        # data fed to the state machine so display_output and the yield
        # shape can be reconstructed after progress_batch() has moved
        # the minion out of ``active``.
        raw_by_minion = {}

        try:
            while not salt.utils.batch_state.is_batch_done(state):
                new_returns, timed_out = self._poll_iterators(
                    iters, minion_tracker, raw_mode, raw_by_minion
                )
                self._discover_late_minions(state)

                now = time.time()
                action = salt.utils.batch_state.progress_batch(
                    state, new_returns, now=now, timed_out=timed_out
                )

                if action.publish:
                    output.on_batch_start(action.publish)
                    args = [
                        list(action.publish),
                        self.opts["fun"],
                        self.opts["arg"],
                        self.opts["timeout"],
                        "list",
                    ]
                    new_iter = self.local.cmd_iter_no_block(
                        *args,
                        raw=raw_mode,
                        ret=return_value,
                        show_jid=show_jid,
                        verbose=show_verbose,
                        gather_job_timeout=self.opts["gather_job_timeout"],
                        jid=batch_jid,
                        **self.eauth,
                    )
                    iters.append(new_iter)
                    minion_tracker[new_iter] = {
                        "minions": list(action.publish),
                        "active": True,
                    }

                halted_mid_yield = False
                for minion_id, data in new_returns.items():
                    if data.get("failed") is True:
                        output.on_minion_failed(minion_id)
                    else:
                        retcode = salt.utils.batch_state._collapse_retcode(data)
                        if raw_mode:
                            yield raw_by_minion.get(minion_id, data), retcode
                        else:
                            yield {minion_id: data.get("ret")}, retcode
                        output.on_minion_return(
                            minion_id, raw_by_minion.get(minion_id, data)
                        )
                    if state["halted"]:
                        halted_mid_yield = True
                        break

                if halted_mid_yield:
                    log.error(
                        "Batch run stopped due to failhard",
                    )
                    break

                for minion_id in timed_out:
                    envelope = raw_by_minion.get(minion_id)
                    if raw_mode and envelope is not None:
                        yield envelope, 0
                    elif raw_mode:
                        yield {"data": {"id": minion_id, "return": {}, "retcode": 0}}, 0
                    else:
                        yield {minion_id: {}}, 0
                    output.on_minion_timeout(minion_id)

                salt.utils.batch_state.write_batch_state(
                    batch_jid, state, self.opts, best_effort=True
                )

                # Prune finished iterators; progress_batch already
                # cleared their minions from state["active"].
                iters = [
                    queue
                    for queue in iters
                    if minion_tracker.get(queue, {}).get("active")
                ]

                # When neither a dispatch nor a poll happened, idle
                # briefly so we don't hot-spin waiting for batch_wait
                # to expire.
                if (
                    not action.publish
                    and not iters
                    and not new_returns
                    and not timed_out
                ):
                    if not salt.utils.batch_state.is_batch_done(state):
                        time.sleep(0.02)

            output.on_batch_done(state)
        finally:
            salt.utils.batch_state.remove_from_active_index(
                batch_jid, self.opts, best_effort=True
            )
            salt.utils.batch_state.write_batch_state(
                batch_jid, state, self.opts, best_effort=True
            )
            self.local.destroy()

    def _poll_iterators(self, iters, minion_tracker, raw_mode, raw_by_minion):
        """
        Drain every active ``cmd_iter_no_block`` iterator once.

        Returns ``(new_returns, timed_out)`` — a dict of normalized
        minion returns and a list of minion IDs the iterator exhausted
        without yielding (i.e. ``cmd_iter_no_block``'s own timeout
        tripped).  ``raw_by_minion`` is populated with the raw
        payloads so the caller can preserve yield shape.

        Iterators that raise ``StopIteration`` are marked inactive in
        ``minion_tracker`` but not removed from ``iters`` — the caller
        filters them after the state-machine step so yields happen in
        iterator order.
        """
        new_returns = {}
        timed_out = []
        for queue in list(iters):
            try:
                ncnt = 0
                while True:
                    part = next(queue)
                    if part is None:
                        time.sleep(0.01)
                        ncnt += 1
                        if ncnt > 5:
                            break
                        continue
                    if raw_mode:
                        minion_id = part["data"]["id"]
                        raw_by_minion[minion_id] = part
                        new_returns[minion_id] = {
                            "ret": part["data"].get("return"),
                            "retcode": part["data"].get("retcode", 0),
                            "failed": part["data"].get("failed", False),
                        }
                        if minion_id in minion_tracker[queue]["minions"]:
                            minion_tracker[queue]["minions"].remove(minion_id)
                        else:
                            if not self.quiet:
                                salt.utils.stringutils.print_cli(
                                    "minion {} was already deleted from tracker,"
                                    " probably a duplicate key".format(minion_id)
                                )
                    else:
                        for minion_id, mret in part.items():
                            raw_by_minion[minion_id] = copy.copy(mret)
                            new_returns[minion_id] = mret
                            if minion_id in minion_tracker[queue]["minions"]:
                                minion_tracker[queue]["minions"].remove(minion_id)
                            else:
                                if not self.quiet:
                                    salt.utils.stringutils.print_cli(
                                        "minion {} was already deleted from tracker,"
                                        " probably a duplicate key".format(minion_id)
                                    )
            except StopIteration:
                if queue in minion_tracker:
                    minion_tracker[queue]["active"] = False
                    for minion_id in minion_tracker[queue]["minions"]:
                        if minion_id not in new_returns:
                            timed_out.append(minion_id)
        return new_returns, timed_out

    def _discover_late_minions(self, state):
        """
        Pull newly discovered minions off ``ping_gen`` and append them
        to ``state["pending"]`` so the state machine will schedule
        them.
        """
        for ping_ret in self.ping_gen:
            if ping_ret is None:
                break
            try:
                minion_id = next(iter(ping_ret.keys()))
            except StopIteration:
                break
            if minion_id not in state["all_minions"]:
                state["all_minions"].append(minion_id)
                state["pending"].append(minion_id)
                if minion_id not in self.minions:
                    self.minions.append(minion_id)



import logging
from io import StringIO

import pytest

import salt.exceptions
import salt.modules.slsutil as slsutil
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules(master_dirs, master_files, minion_opts):
    return {
        slsutil: {
            "__opts__": minion_opts,
            "__salt__": {
                "cp.list_master": MagicMock(return_value=master_files),
                "cp.list_master_dirs": MagicMock(return_value=master_dirs),
            },
        }
    }


@pytest.fixture
def master_dirs():
    return ["red", "red/files", "blue", "blue/files"]


@pytest.fixture
def master_files():
    return [
        "top.sls",
        "red/init.sls",
        "red/files/default.conf",
        "blue/init.sls",
        "blue/files/default.conf",
    ]


def test_banner():
    """
    Test banner function
    """
    check_banner()
    check_banner(width=81)
    check_banner(width=20)
    check_banner(commentchar="//", borderchar="-")
    check_banner(title="title here", text="text here")
    check_banner(commentchar=" *")
    check_banner(commentchar=" *", newline=False)

    # Test when width result in a raised exception
    with pytest.raises(salt.exceptions.ArgumentValueError):
        slsutil.banner(width=4)

    ret = slsutil.banner(
        title="title here", text="text here", blockstart="/*", blockend="*/"
    )
    lines = ret.splitlines()
    # test blockstart
    assert lines[0] == "/*"

    # test blockend
    assert lines[-1] == "*/"


def check_banner(
    width=72,
    commentchar="#",
    borderchar="#",
    blockstart=None,
    blockend=None,
    title=None,
    text=None,
    newline=True,
):

    result = slsutil.banner(
        width=width,
        commentchar=commentchar,
        borderchar=borderchar,
        blockstart=blockstart,
        blockend=blockend,
        title=title,
        text=text,
        newline=newline,
    ).splitlines()
    for line in result:
        assert len(line) == width
        assert line.startswith(commentchar)
        assert line.endswith(commentchar.strip())


def test_boolstr():
    """
    Test boolstr function
    """
    assert "yes" == slsutil.boolstr(True, true="yes", false="no")
    assert "no" == slsutil.boolstr(False, true="yes", false="no")


def test_file_exists():
    """
    Test file_exists function
    """
    assert slsutil.file_exists("red/init.sls")
    assert not slsutil.file_exists("green/init.sls")


def test_dir_exists():
    """
    Test dir_exists function
    """
    assert slsutil.dir_exists("red")
    assert not slsutil.dir_exists("green")


def test_path_exists():
    """
    Test path_exists function
    """
    assert slsutil.path_exists("red")
    assert not slsutil.path_exists("green")
    assert slsutil.path_exists("red/init.sls")
    assert not slsutil.path_exists("green/init.sls")


def test_findup():
    """
    Test findup function
    """
    assert "red/init.sls" == slsutil.findup("red/files", "init.sls")
    assert "top.sls" == slsutil.findup("red/files", ["top.sls"])
    assert "top.sls" == slsutil.findup("", "top.sls")
    assert "top.sls" == slsutil.findup(None, "top.sls")
    assert "red/init.sls" == slsutil.findup("red/files", ["top.sls", "init.sls"])

    with pytest.raises(salt.exceptions.CommandExecutionError):
        slsutil.findup("red/files", "notfound")

    with pytest.raises(salt.exceptions.CommandExecutionError):
        slsutil.findup("red", "default.conf")

    with pytest.raises(salt.exceptions.SaltInvocationError):
        with patch.object(slsutil, "path_exists", return_value=False):
            slsutil.findup("red", "default.conf")

    with pytest.raises(salt.exceptions.SaltInvocationError):
        slsutil.findup("red", {"file": "default.conf"})


def test_update():
    """
    Test update function
    """

    ret = slsutil.update({"foo": "Foo"}, {"bar": "Bar"})
    assert ret == {"foo": "Foo", "bar": "Bar"}

    ret = slsutil.update({"foo": "Foo"}, {"foo": "Bar"}, merge_lists=False)
    assert ret == {"foo": "Bar"}


def test_merge():
    """
    Test merge function
    """

    ret = slsutil.merge({"foo": "Foo"}, {"bar": "Bar"}, strategy="smart")
    assert ret == {"foo": "Foo", "bar": "Bar"}

    ret = slsutil.merge({"foo": "Foo"}, {"foo": "Bar"}, strategy="aggregate")
    assert ret == {"foo": "Bar"}

    ret = slsutil.merge({"foo": "Foo"}, {"foo": "Bar"}, strategy="list")
    assert ret == {"foo": ["Foo", "Bar"]}

    ret = slsutil.merge({"foo": "Foo"}, {"foo": "Bar"}, strategy="overwrite")
    assert ret == {"foo": "Bar"}

    ret = slsutil.merge(
        {"foo": {"Foo": "Bar"}}, {"foo": {"Foo": "Baz"}}, strategy="recurse"
    )
    assert ret == {"foo": {"Foo": "Baz"}}


def test_merge_all():
    """
    Test merge_all function
    """

    ret = slsutil.merge_all([{"foo": "Foo"}, {"bar": "Bar"}], strategy="smart")
    assert ret == {"foo": "Foo", "bar": "Bar"}

    ret = slsutil.merge_all([{"foo": "Foo"}, {"foo": "Bar"}], strategy="aggregate")
    assert ret == {"foo": "Bar"}

    ret = slsutil.merge_all([{"foo": "Foo"}, {"foo": "Bar"}], strategy="overwrite")
    assert ret == {"foo": "Bar"}

    ret = slsutil.merge_all(
        [{"foo": {"Foo": "Bar"}}, {"foo": {"Foo": "Baz"}}], strategy="recurse"
    )
    assert ret == {"foo": {"Foo": "Baz"}}


def test_renderer():
    """
    Test renderer function
    """
    with patch.dict(
        slsutil.__utils__, {"stringio.is_readable": MagicMock(return_value=False)}
    ):
        ret = slsutil.renderer(string="Hello, {{ name }}.", name="world")
        assert ret == "Hello, world."

    with pytest.raises(salt.exceptions.SaltInvocationError) as exc:
        slsutil.renderer()
    assert str(exc.value) == "Must pass path or string."

    with pytest.raises(salt.exceptions.SaltInvocationError) as exc:
        slsutil.renderer(path="/path/to/file", string="Hello world")
    assert str(exc.value) == "Must not pass both path and string."

    with patch.dict(
        slsutil.__salt__, {"cp.get_url": MagicMock(return_value="/path/to/file")}
    ):
        with patch.dict(
            slsutil.__utils__, {"stringio.is_readable": MagicMock(return_value=True)}
        ):
            rendered_file = "Hello, world."
            with patch(
                "salt.template.compile_template",
                MagicMock(return_value=StringIO(rendered_file)),
            ):
                ret = slsutil.renderer(path="/path/to/file")
                assert ret == "Hello, world."


def test_serialize():
    """
    Test serialize function
    """
    ret = slsutil.serialize("json", obj={"foo": "Foo!"})
    assert ret == '{"foo": "Foo!"}'


def test_deserialize():
    """
    Test serialize function
    """
    ret = slsutil.deserialize("json", '{"foo": "Foo!"}')
    assert ret == {"foo": "Foo!"}


def dummy_function(args=None, kwargs=None):
    return True


def test__set_context():
    """
    Test _set_context
    """
    with patch.dict(slsutil.__context__, {}):

        slsutil._set_context(
            ["level_one", "level_two", "level_three"], dummy_function, force=True
        )
        assert slsutil.__context__ == {
            "level_one": {"level_two": {"level_three": True}}
        }

    with patch.dict(slsutil.__context__, {}):

        slsutil._set_context(
            ["level_one", "level_two", "level_three"],
            dummy_function,
            fun_kwargs={"key_one": "arg_one"},
            force=True,
        )
        assert slsutil.__context__ == {
            "level_one": {"level_two": {"level_three": True}}
        }


def test__get_serializer_fn():
    """
    Test _set_context
    """
    # Invalid serializer
    with pytest.raises(salt.exceptions.CommandExecutionError) as exc:
        slsutil._get_serialize_fn("bad_yaml", "badfunc")
        assert str(exc.value) == "Serializer 'bad_yaml' not found."

    # Invalid serializer function
    with pytest.raises(salt.exceptions.CommandExecutionError) as exc:
        slsutil._get_serialize_fn("yaml", "foobar")
        assert str(exc.value) == "Serializer 'yaml' does not implement foobar."



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.cmd as cmd
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {cmd: {"__env__": "base"}}


# 'wait' function tests: 1


def test_wait():
    """
    Test to run the given command only if the watch statement calls it.
    """
    name = "cmd.script"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    assert cmd.wait(name) == ret


# 'wait_script' function tests: 1


def test_wait_script():
    """
    Test to download a script from a remote source and execute it
    only if a watch statement calls it.
    """
    name = "cmd.script"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    assert cmd.wait_script(name) == ret


# 'run' function tests: 1


def test_run():
    """
    Test to run a command if certain circumstances are met.
    """
    name = "cmd.script"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    with patch.dict(cmd.__opts__, {"test": False}):
        comt = "Invalidly-formatted 'env' parameter. See documentation."
        ret.update({"comment": comt})
        assert cmd.run(name, env="salt") == ret

    with patch.dict(cmd.__grains__, {"shell": "shell"}):
        with patch.dict(cmd.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[CommandExecutionError, {"retcode": 1}])
            with patch.dict(cmd.__salt__, {"cmd.run_all": mock}):
                ret.update({"comment": "", "result": False})
                assert cmd.run(name) == ret

                ret.update(
                    {
                        "comment": 'Command "cmd.script" run',
                        "result": False,
                        "changes": {"retcode": 1},
                    }
                )
                assert cmd.run(name) == ret

        with patch.dict(cmd.__opts__, {"test": True}):
            comt = 'Command "cmd.script" would have been executed'
            ret.update(
                {"comment": comt, "result": None, "changes": {"cmd": "cmd.script"}}
            )
            assert cmd.run(name) == ret


def test_run_root():
    """
    Test to run a command with a different root
    """
    name = "cmd.script"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    with patch.dict(cmd.__grains__, {"shell": "shell"}):
        with patch.dict(cmd.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[CommandExecutionError, {"retcode": 1}])
            with patch.dict(cmd.__salt__, {"cmd.run_chroot": mock}):
                ret.update({"comment": "", "result": False})
                assert cmd.run(name, root="/mnt") == ret

                ret.update(
                    {
                        "comment": 'Command "cmd.script" run',
                        "result": False,
                        "changes": {"retcode": 1},
                    }
                )
                assert cmd.run(name, root="/mnt") == ret


# 'script' function tests: 1


def test_script():
    """
    Test to download a script and execute it with specified arguments.
    """
    name = "cmd.script"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    with patch.dict(cmd.__opts__, {"test": False}):
        comt = "Invalidly-formatted 'env' parameter. See documentation."
        ret.update({"comment": comt})
        assert cmd.script(name, env="salt") == ret

    with patch.dict(cmd.__grains__, {"shell": "shell"}):
        with patch.dict(cmd.__opts__, {"test": True}):
            comt = "Command 'cmd.script' would have been executed"
            ret.update({"comment": comt, "result": None, "changes": {}})
            assert cmd.script(name) == ret

        with patch.dict(cmd.__opts__, {"test": False}):
            mock = MagicMock(side_effect=[CommandExecutionError, {"retcode": 1}])
            with patch.dict(cmd.__salt__, {"cmd.script": mock}):
                ret.update({"comment": "", "result": False})
                assert cmd.script(name) == ret

                ret.update(
                    {
                        "comment": "Command 'cmd.script' run",
                        "result": False,
                        "changes": {"retcode": 1},
                    }
                )
                assert cmd.script(name) == ret


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
def test_script_runas_no_password():
    """
    Test to download a script and execute it, specifically on windows.
    Make sure that when using runas, and a password is not supplied we check.
    """
    name = "cmd.script"
    ret = {
        "name": "cmd.script",
        "changes": {},
        "result": False,
        "comment": "Must supply a password if runas argument is used on Windows.",
    }

    patch_opts = patch.dict(cmd.__opts__, {"test": False})
    patch_shell = patch.dict(cmd.__grains__, {"shell": "shell"})

    # Fake the execution module call
    mock = MagicMock(side_effect=[CommandExecutionError, {"retcode": 1}])
    patch_call = patch.dict(cmd.__salt__, {"cmd.script": mock})

    with patch_opts, patch_shell, patch_call:
        assert cmd.script(name, runas="User") == ret


@pytest.mark.skip_unless_on_windows(reason="Test is only applicable to Windows.")
def test_script_runas_password():
    """
    Test to download a script and execute it, specifically on windows.
    Make sure that when we do supply a password, we get into the function call.
    """
    name = "cmd.script"
    ret = {"changes": {}, "comment": "", "name": "cmd.script", "result": False}

    patch_opts = patch.dict(cmd.__opts__, {"test": False})
    patch_shell = patch.dict(cmd.__grains__, {"shell": "shell"})

    # Fake the execution module call
    mock = MagicMock(side_effect=[CommandExecutionError, {"retcode": 1}])
    patch_call = patch.dict(cmd.__salt__, {"cmd.script": mock})

    with patch_opts, patch_shell, patch_call:
        assert cmd.script(name, runas="User", password="Test") == ret


# 'call' function tests: 1


def test_call():
    """
    Test to invoke a pre-defined Python function with arguments
    specified in the state declaration.
    """
    name = "cmd.script"
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    flag = None

    def func():
        """
        Mock func method
        """
        if flag:
            return {}
        else:
            return []

    with patch.dict(cmd.__grains__, {"shell": "shell"}):
        flag = True
        assert cmd.call(name, func) == ret

        flag = False
        ret.update({"comment": "", "result": False, "changes": {"retval": []}})
        assert cmd.call(name, func) == ret


# 'wait_call' function tests: 1


def test_wait_call():
    """
    Test to run wait_call.
    """
    name = "cmd.script"
    func = "myfunc"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    assert cmd.wait_call(name, func) == ret


# 'mod_watch' function tests: 1


def test_mod_watch():
    """
    Test to execute a cmd function based on a watch call
    """
    name = "cmd.script"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    def func():
        """
        Mock func method
        """
        return {}

    with patch.dict(cmd.__grains__, {"shell": "shell"}):
        with patch.dict(cmd.__opts__, {"test": False}):
            mock = MagicMock(return_value={"retcode": 1})
            with patch.dict(cmd.__salt__, {"cmd.run_all": mock}):
                assert cmd.mod_watch(name, sfun="wait", stateful=True) == ret

                comt = 'Command "cmd.script" run'
                ret.update({"comment": comt, "changes": {"retcode": 1}})
                assert cmd.mod_watch(name, sfun="wait", stateful=False) == ret

            with patch.dict(cmd.__salt__, {"cmd.script": mock}):
                ret.update({"comment": "", "changes": {}})
                assert cmd.mod_watch(name, sfun="script", stateful=True) == ret

                comt = "Command 'cmd.script' run"
                ret.update({"comment": comt, "changes": {"retcode": 1}})
                assert cmd.mod_watch(name, sfun="script", stateful=False) == ret

            with patch.dict(cmd.__salt__, {"cmd.script": mock}):
                ret.update({"comment": "", "changes": {}})
                assert cmd.mod_watch(name, sfun="call", func=func) == ret

                comt = "cmd.call needs a named parameter func"
                ret.update({"comment": comt})
                assert cmd.mod_watch(name, sfun="call") == ret

            comt = (
                "cmd.salt does not work with the watch requisite,"
                " please use cmd.wait or cmd.wait_script"
            )
            ret.update({"comment": comt})
            assert cmd.mod_watch(name, sfun="salt") == ret



"""
Writing/reading defaults from a macOS minion
============================================

"""

import salt.utils.platform

__virtualname__ = "macdefaults"


def __virtual__():
    """
    Only work on macOS
    """
    if salt.utils.platform.is_darwin():
        return __virtualname__
    return (False, "Only supported on macOS")


def write(name, domain, value, vtype=None, name_separator=None, user=None):
    """
    Write a default to the system

    name
        The key of the given domain to write to.
        It can be a nested key/index separated by `name_separator`.

    name_separator
        The separator to use when splitting the name into a list of keys.
        If None, the name will not be split (Default).

        .. versionadded:: 3008.0

    domain
        The name of the domain to write to

    value
        The value to write to the given key

    vtype
        The type of value to be written. Valid types are string, data, int[eger],
        float, bool[ean], date, array, array-add, dict, dict-add

    user
        The user to write the defaults to

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    current_value = __salt__["macdefaults.read"](
        domain, name, user, key_separator=name_separator
    )

    if vtype is not None:
        try:
            value = __salt__["macdefaults.cast_value_to_vtype"](value, vtype)
        except ValueError as e:
            ret["result"] = False
            ret["comment"] = f"Failed to cast value {value} to {vtype}. {e}"
            return ret

    if _compare_values(value, current_value, vtype):
        ret["comment"] = f"{domain} {name} is already set to {value}"
        return ret

    try:
        out = __salt__["macdefaults.write"](
            domain, name, value, vtype, user, key_separator=name_separator
        )
        if out["retcode"] != 0:
            ret["result"] = False
            ret["comment"] = f"Failed to write default. {out['stderr']}"
        else:
            ret["changes"]["written"] = f"{domain} {name} is set to {value}"
    # pylint: disable-next=broad-exception-caught
    except Exception as e:
        ret["result"] = False
        ret["comment"] = f"Failed to write default. {e}"

    return ret


def absent(name, domain, user=None, name_separator=None):
    """
    Make sure the defaults value is absent

    name
        The key of the given domain to remove.
        It can be a nested key/index separated by `name_separator`.

    name_separator
        The separator to use when splitting the name into a list of keys.
        If None, the name will not be split (Default).

        .. versionadded:: 3008.0

    domain
        The name of the domain to remove from

    user
        The user to write the defaults to

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    out = __salt__["macdefaults.delete"](
        domain, name, user, key_separator=name_separator
    )

    if out is None or out["retcode"] != 0:
        ret["comment"] += f"{domain} {name} is already absent"
    else:
        ret["changes"]["absent"] = f"{domain} {name} is now absent"

    return ret


def _compare_values(new, current, vtype):
    """
    Compare two values based on their type

    new
        The new value to compare

    current
        The current value to compare

    vtype
        The type of default value to be compared

    """
    if vtype == "array-add":
        if isinstance(new, list):
            return new == current[-len(new) :]
        return new == current[-1]

    if vtype == "dict-add":
        return all(key in current and new[key] == current[key] for key in new.keys())

    return new == current



import pytest
from salt.utils.templates import render_cheetah_tmpl

pytest.importorskip("Cheetah")


def test_render_sanity(render_context):
    tmpl = """OK"""
    res = render_cheetah_tmpl(tmpl, render_context)
    assert res == "OK"


def test_render_evaluate(render_context):
    tmpl = """<%="OK"%>"""
    res = render_cheetah_tmpl(tmpl, render_context)
    assert res == "OK"


def test_render_evaluate_xml(render_context):
    tmpl = """
    <% if 1: %>
    OK
    <% pass %>
    """
    res = render_cheetah_tmpl(tmpl, render_context)
    stripped = res.strip()
    assert stripped == "OK"


def test_render_evaluate_text(render_context):
    tmpl = """
    #if 1
    OK
    #end if
    """

    res = render_cheetah_tmpl(tmpl, render_context)
    stripped = res.strip()
    assert stripped == "OK"


def test_render_variable(render_context):
    tmpl = """$var"""

    render_context["var"] = "OK"
    res = render_cheetah_tmpl(tmpl, render_context)
    assert res.strip() == "OK"



from textwrap import dedent

import pytest

import salt.utils.x509 as x509

cx509 = pytest.importorskip("cryptography.x509")


@pytest.fixture
def b64cert_with_prefix():
    return (
        "b64:MIIF6jCCA9KgAwIBAgIUHkYQ5opY8AXgK7RNSqUtMcltnqMwDQYJKoZIhvcNAQELBQAwSTELMAkGA1UEBhMCVV"
        "MxCzAJBgNVBAgMAk1EMRMwEQYDVQQHDApTeWtlc3ZpbGxlMRgwFgYDVQQDDA9jYS5jZHguZWl0ci5kZXYwHhcNMjQw"
        "MzI3MTg0MzU0WhcNMjQwNDI2MTg0MzU0WjBLMQswCQYDVQQGEwJVUzELMAkGA1UECAwCTUQxEzARBgNVBAcMClN5a2"
        "VzdmlsbGUxGjAYBgNVBAMMEW5pZmkuY2R4LmVpdHIuZGV2MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA"
        "zEhNiCogpOdh6kK+wkh+rBe8/zyE6O0XjcWaEm+i/dhG35KU/c6zZhmkNObtrEwvrqIIKpca2h3IaRb6FAp2VpedGy"
        "4/bVihEVRymZOtGo8Yex74THmokkngTfnxyfyZdULc7YL7Pi/FPejcCy8lWypcnLzpTnw0qx2GmRmENyrXvqrB429L"
        "HzefZv/FCDPZixqkUuaK3iPqhJd83HXb9BOyi8BtF6b7qrnds0KlivIO/zCUZnfOn2610Dja82eSFASkgDbNJsJn37"
        "ktEhbHGtkkCVD6zBH0p0dgXnjQ8Ml0+QJIoSl8RBe2EkZ0ZIMKHIOfleOBOI6Cd2CYyDWjRxD3nFqcRnNGhLNBspm8"
        "s8C+3e1iyZQ224fy6BA5FHp3M0UX6ct1+M3JzxxLAbSuG8pc4MC3DLGDK4OlLbAnpFYqBAALs5OKTptxU4eEZqdFfj"
        "9PFNknU1lFVrqGFbaE/oRrORsznNFZm3gxRSIvNtDuBJOYUl4KsYHjOjM/G3jRzc1+1K7wVpMoO/kdjIo2zhMEbBTw"
        "Lx0xrgBQzzVLLmsib4cFts8zELFkB5nGl1mv2+KSOjQ+gpQtn0lkYSY7iVfVSt13JRY7mIOTnmjHj5mRguvgbr3dNa"
        "VfQMCJD7pOMBaxO5O0aiwVE8KjNz9WEDqrzW0BG+ei3fLosDIvbIkCAwEAAaOBxzCBxDAMBgNVHRMBAf8EAjAAMA4G"
        "A1UdDwEB/wQEAwIFIDAdBgNVHQ4EFgQUTOTqSBdqbMm4lLxIupUhsTeYPXMwgYQGA1UdIwR9MHuAFBN3hzb/2SCZZl"
        "BiHUIZYTJXQZIMoU2kSzBJMQswCQYDVQQGEwJVUzELMAkGA1UECAwCTUQxEzARBgNVBAcMClN5a2VzdmlsbGUxGDAW"
        "BgNVBAMMD2NhLmNkeC5laXRyLmRldoIUDVzffz0J8C716U6jXZszcredC1owDQYJKoZIhvcNAQELBQADggIBAGSS/d"
        "iai+Imm2559MzTYK5qvCVWCDaizAgH6JZeLZGf9Mk7IEZrS3I9UtjnVH9q4VON5KJtz+CvYU/t+el0AsEfns8Tw/Ff"
        "MBTD7cBFBBPtIPxpYh0nzpEvxI8sxKkFt1vmDMuYiBGkPx1OTLwTbL6EbAJznooiWIg0n59Wd1Jn3U8Q4O6/yLy23x"
        "ZA/xUSjgIbTXOctBzYC47FwNyjcaQ70gLZJC/pCd+hUoojBaAUHNfuzK0RqF7eP6W67nGVyA1h/B87FG0y6tmuRWWl"
        "jwyAz/Nvjb2SXWkgxxkS4ZPZt6z+R8FsRSbMuIR5CeOyMeKUbQfc3hWvII9c7mZkZRYnxUuFqpwUlOWnNX1ufikBQE"
        "OOyta3n/Lbj59+QBmPU8ok+RBfyCEKDVw5DAhu95gj6rdxUeWrGLteR8o0O/n6JGnM0B5kJ7y2NnaLa06QYzJUmSs5"
        "/icBRwyGSL3Gw9GkkRpGNViRIMpcrqGvr5bYxFeNkQGqiB+0vxiD6s1DOz7djY4K03ZUGYLe3X73CKu+AxbhC95sz6"
        "hWURdotqO4CUb9Nd82sY2HCDBFPEFnT1RD+Xi6nkULvHkquhYVV3eHC4LtvhlHjF1LufZ7xOYoteScZL5WvumvrdNS"
        "9naI8BZkWtsTl98Z2GhuZPKpOQtMOPXC38qEuNc5UPJhb3Oa"
    )


@pytest.fixture
def b64cert(b64cert_with_prefix):
    return b64cert_with_prefix[4:]


@pytest.fixture
def pemcert():
    return dedent(
        """-----BEGIN CERTIFICATE-----
        MIIF6jCCA9KgAwIBAgIUHkYQ5opY8AXgK7RNSqUtMcltnqMwDQYJKoZIhvcNAQEL
        BQAwSTELMAkGA1UEBhMCVVMxCzAJBgNVBAgMAk1EMRMwEQYDVQQHDApTeWtlc3Zp
        bGxlMRgwFgYDVQQDDA9jYS5jZHguZWl0ci5kZXYwHhcNMjQwMzI3MTg0MzU0WhcN
        MjQwNDI2MTg0MzU0WjBLMQswCQYDVQQGEwJVUzELMAkGA1UECAwCTUQxEzARBgNV
        BAcMClN5a2VzdmlsbGUxGjAYBgNVBAMMEW5pZmkuY2R4LmVpdHIuZGV2MIICIjAN
        BgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAzEhNiCogpOdh6kK+wkh+rBe8/zyE
        6O0XjcWaEm+i/dhG35KU/c6zZhmkNObtrEwvrqIIKpca2h3IaRb6FAp2VpedGy4/
        bVihEVRymZOtGo8Yex74THmokkngTfnxyfyZdULc7YL7Pi/FPejcCy8lWypcnLzp
        Tnw0qx2GmRmENyrXvqrB429LHzefZv/FCDPZixqkUuaK3iPqhJd83HXb9BOyi8Bt
        F6b7qrnds0KlivIO/zCUZnfOn2610Dja82eSFASkgDbNJsJn37ktEhbHGtkkCVD6
        zBH0p0dgXnjQ8Ml0+QJIoSl8RBe2EkZ0ZIMKHIOfleOBOI6Cd2CYyDWjRxD3nFqc
        RnNGhLNBspm8s8C+3e1iyZQ224fy6BA5FHp3M0UX6ct1+M3JzxxLAbSuG8pc4MC3
        DLGDK4OlLbAnpFYqBAALs5OKTptxU4eEZqdFfj9PFNknU1lFVrqGFbaE/oRrORsz
        nNFZm3gxRSIvNtDuBJOYUl4KsYHjOjM/G3jRzc1+1K7wVpMoO/kdjIo2zhMEbBTw
        Lx0xrgBQzzVLLmsib4cFts8zELFkB5nGl1mv2+KSOjQ+gpQtn0lkYSY7iVfVSt13
        JRY7mIOTnmjHj5mRguvgbr3dNaVfQMCJD7pOMBaxO5O0aiwVE8KjNz9WEDqrzW0B
        G+ei3fLosDIvbIkCAwEAAaOBxzCBxDAMBgNVHRMBAf8EAjAAMA4GA1UdDwEB/wQE
        AwIFIDAdBgNVHQ4EFgQUTOTqSBdqbMm4lLxIupUhsTeYPXMwgYQGA1UdIwR9MHuA
        FBN3hzb/2SCZZlBiHUIZYTJXQZIMoU2kSzBJMQswCQYDVQQGEwJVUzELMAkGA1UE
        CAwCTUQxEzARBgNVBAcMClN5a2VzdmlsbGUxGDAWBgNVBAMMD2NhLmNkeC5laXRy
        LmRldoIUDVzffz0J8C716U6jXZszcredC1owDQYJKoZIhvcNAQELBQADggIBAGSS
        /diai+Imm2559MzTYK5qvCVWCDaizAgH6JZeLZGf9Mk7IEZrS3I9UtjnVH9q4VON
        5KJtz+CvYU/t+el0AsEfns8Tw/FfMBTD7cBFBBPtIPxpYh0nzpEvxI8sxKkFt1vm
        DMuYiBGkPx1OTLwTbL6EbAJznooiWIg0n59Wd1Jn3U8Q4O6/yLy23xZA/xUSjgIb
        TXOctBzYC47FwNyjcaQ70gLZJC/pCd+hUoojBaAUHNfuzK0RqF7eP6W67nGVyA1h
        /B87FG0y6tmuRWWljwyAz/Nvjb2SXWkgxxkS4ZPZt6z+R8FsRSbMuIR5CeOyMeKU
        bQfc3hWvII9c7mZkZRYnxUuFqpwUlOWnNX1ufikBQEOOyta3n/Lbj59+QBmPU8ok
        +RBfyCEKDVw5DAhu95gj6rdxUeWrGLteR8o0O/n6JGnM0B5kJ7y2NnaLa06QYzJU
        mSs5/icBRwyGSL3Gw9GkkRpGNViRIMpcrqGvr5bYxFeNkQGqiB+0vxiD6s1DOz7d
        jY4K03ZUGYLe3X73CKu+AxbhC95sz6hWURdotqO4CUb9Nd82sY2HCDBFPEFnT1RD
        +Xi6nkULvHkquhYVV3eHC4LtvhlHjF1LufZ7xOYoteScZL5WvumvrdNS9naI8BZk
        WtsTl98Z2GhuZPKpOQtMOPXC38qEuNc5UPJhb3Oa
        -----END CERTIFICATE-----"""
    )


def test_load_file_or_bytes_base64_der_with_b64_prefix(b64cert_with_prefix):
    der = x509.load_file_or_bytes(b64cert_with_prefix)
    cert = cx509.load_der_x509_certificate(der)
    assert (
        cert.subject.rfc4514_string() == "CN=nifi.cdx.eitr.dev,L=Sykesville,ST=MD,C=US"
    )


def test_load_file_or_bytes_base64_der(b64cert):
    der = x509.load_file_or_bytes(b64cert)
    cert = cx509.load_der_x509_certificate(der)
    assert (
        cert.subject.rfc4514_string() == "CN=nifi.cdx.eitr.dev,L=Sykesville,ST=MD,C=US"
    )


def test_load_file_or_bytes_pem(pemcert):
    pem = x509.load_file_or_bytes(pemcert)
    cert = cx509.load_pem_x509_certificate(pem)
    assert (
        cert.subject.rfc4514_string() == "CN=nifi.cdx.eitr.dev,L=Sykesville,ST=MD,C=US"
    )



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_pyobjects_renderer(state, state_tree, tmp_path):
    """
    Test pyobjects renderer when running state.sls
    """
    file_path = str(tmp_path).replace("\\", "/")
    sls1_contents = f"""
    #!pyobjects
    import pathlib
    import salt://test_pyobjects2.sls
    test_file = pathlib.Path("{file_path}", "test")
    File.managed(str(test_file))
    """
    sls2_contents = f"""
    #!pyobjects
    import pathlib
    test_file = pathlib.Path("{file_path}", "test2")
    File.managed(str(test_file))
    """

    with pytest.helpers.temp_file(
        "test_pyobjects.sls", sls1_contents, state_tree
    ) as state1:
        with pytest.helpers.temp_file(
            "test_pyobjects2.sls", sls2_contents, state_tree
        ) as state2:
            ret = state.sls("test_pyobjects")
            assert not ret.errors
            for state_return in ret:
                assert state_return.result is True
                assert str(tmp_path) in state_return.name
